package		MakeParagraph;
require		Exporter;
@ISA =		qw(Exporter);
@EXPORT =	qw(new clear empty string width set_width append leave
		simple_wrap white_space white_space_after_most format);

use strict;
use English;
use Text::Wrap;

# (C) 1999 David C. Fox
# This code may be freely reproduced and modified, provided 
# 1. This copyright notice is retained, 
# and
# 2. substantial modifications are indicated, together with the author's
# name, in any redistributed versions.


sub new {
  my $invoked = shift;
  my $self;
  my $type = ref( $invoked);
  if ( $type) {
    my %temporary = %$invoked;
    $self = \%temporary;
    bless $self, $type;
  } else {
    $self = {};
    bless $self, $invoked;
    $self->_initialize(@_);
  }
  return $self;
}

sub _initialize {
  my $self = shift;
  my $width = shift;
  $self->{"string"} = "";
  if ( $width) {
    $self->{"line width"} = $width;
  } else {
    $self->{"line width"} = 72;
  }
  return 1;
}

sub clear {
  my $self = shift;
  $self->{"string"} = "";
  return;
}


sub empty {
  my $self = shift;
  return ($self->string() eq "");
}


sub width {
  my $self = shift;
  return $self->{"line width"};
}

sub string {
  my $self = shift;
  return $self->{"string"};
}

sub set_width {
  my $self = shift;
  my $width = shift;
  if ( $width) {
    $self->{"line width"} = $width;
  }
  return;
}

sub leave {
  my $self = shift;
  return $self->MakeParagraph::append( @_);
}

sub append {
  my $self = shift;
  my $additional = shift;
  $self->{"string"} .= $additional;
  return;
}

sub white_space {
#  append white space (or any other character) as long as the previous
#  character is not itself white space
#  otherwise do nothing
  my $self = shift;
  my $additional = shift;
  if (  !(defined($additional))) {
    $additional = " "; #  default is a single space
  }
  if ( $self->empty()) {
    return 0;
  }

  if ( $self->string() =~ /\s$/) {
    return 0;
  }
  $self->append($additional);
  return 1;
}

sub white_space_after_most {
#  append white space (or any other character) as long as the previous
#  character is not itself white space, or ( or { or [ or ",
#  otherwise do nothing
  my $self = shift;
  my $additional = shift;
  if (  !(defined($additional))) {
    $additional = " "; #  default is a single space
  }
  if ( $self->empty()) {
    return 0;
  }
  if ( $self->string() =~ /[\s\(\{\[\"]$/) {
    return 0;
  }
  $self->append($additional);
  return 1;
}

sub format {
  my $self = shift;
#   $Text::Wrap::columns = $self->width();
#   return wrap("", "", $self->string());
  my $w = $self->simple_wrap();
  chomp $w;
  return $w;
}

sub simple_wrap {
  my $self = shift;
  my $original = shift;
  if ( !(defined ($original))) {
    $original = $self->string();
  }

  my $output = "";
  my $this_line;
  my $width = $self->width();

  my $line;

#     $line = substr( $original, 0, $width);
#     print "starting to process paragraph beginning with\n";
#     print $line . "\n";
  while ( length($original) >= $width) {
    $line = substr( $original, 0, $width);
#     print "[". $line . "]\n";
    if ( $line =~ s/^.*?\n//) {
      $this_line = $&;
      chop $this_line;
    } elsif ( $line =~ s/^.* //) {
      $this_line = $&;
      chop $this_line;
    } else {
      $this_line = $line;
      chop $this_line;
    }
    my $used = length($this_line)+1;
    $original = substr( $original, $used);
    $output .= $this_line . "\n";
  }
  while ( length($original)) {
    if ( $original =~ s/^.*?\n//) {
      $this_line = $&;
      $output .= $this_line;
    } else {
      $output .= $original."\n";
      last;
    }
  }
  return $output;
}

