package		NLXParagraph;
require		Exporter;
use		MakeParagraph();
@ISA =		qw(MakeParagraph);
@EXPORT =	qw(new equation_mode append text_append leave
		math_append white_space_after_most clear
		white_space_after_most_math wrapped unwrapped
		flush_paragraph);

use strict;
use English;
# (C) 1999 David C. Fox
# This code may be freely reproduced and modified, provided 
# 1. This copyright notice is retained, 
# and
# 2. substantial modifications are indicated, together with the author's
# name, in any redistributed versions.

sub new {
  my $invoked = shift;
  my $equation = shift;
  if ( !defined($equation)) {
    $equation = 0;
  }
  my $type = ref( $invoked) || $invoked;
  my $self = new MakeParagraph (@_);
  if ( ref( $self)) {
    #  valid reference returned by MakeParagraph::new, so bless it as type
    #  (either NLXParagraph or a descendant) and then do our initialization
    bless $self, $type;
    $self->equation_mode($equation);
    $self->wrapped(1);
    return $self;
  }
  return $self;
}

sub clear {
  my $self = shift;
  $self->SUPER::clear();
  $self->wrapped(1);
}

sub wrapped {
  my $self = shift;
  my $value = shift;
  if ( defined ($value)) {
    $self->{"wrapped"} = $value;
  }

  return $self->{"wrapped"};
}

sub unwrapped {
  my $self = shift;
  return $self->wrapped(0);
}



sub equation_mode {
#  returns the current setting of the equation mode flag.  If this flag
#  is set, all calls to NLXParagraph::append will be treated as calls to
#  math_append.
#  To set the value of equation mode, simply call equation_mode with an
#  argument.
  my $self = shift;
  my $new_value = shift;
  if ( defined($new_value)) {
    $self->{"equation mode"} = $new_value;
  }
  return $self->{"equation mode"};
}

sub leave {
  my $self = shift;
  return $self->append(@_);
}

sub append {
  my $self = shift;
  if ( $self->equation_mode()) {
    return $self->math_append(@_);
  }
  $self->text_append(@_);
}

sub text_append {
  my $self = shift;
  return $self->SUPER::append(@_);
}

sub math_append {
  my $self = shift;
  my $additional = shift;
  $additional =~ s/^\s+//;
  $additional =~ s/\s+$//;
  if ( ($additional ne "") && !($additional =~ /^[\.\^_]/)) {
    $self->white_space_after_most_math();
  }
  $self->SUPER::append($additional);
  return;
}

sub white_space_after_most_math {
#  append white space (or any other character) as long as the previous
#  character is not itself white space, or $ or _ or ^ or .,
#  otherwise do nothing
  my $self = shift;
  my $additional = shift;
  if (  !(defined($additional))) {
    $additional = " "; #  default is a single space
  }
  if ( $self->empty()) {
    return 0;
  }

  if ( $self->string() =~ /[\.\s\$_\^]$/) {
    return 0;
  }
  $self->SUPER::append($additional);
  return 1;
}

sub white_space_after_most {
#  append white space (or any other character) as long as the previous
#  character is not itself white space, or " or _ or ^ or .,
#  otherwise do nothing
  my $self = shift;
  my $additional = shift;
  if (  !(defined($additional))) {
    $additional = " "; #  default is a single space
  }
  if ( $self->empty()) {
    return 0;
  }

  if ( $self->string() =~ /[\.\s\"_\^]$/) {
    return 0;
  }
  $self->SUPER::append($additional);
  return 1;
}

sub flush_paragraph {
  my $self = shift;
  my ($output_file_ref, $new_lines) = @_;
  my $term = "\n\n";
  if ( defined($new_lines) ) {
    $term = "\n" x $new_lines;
  }
  if ( ! ($self->empty())) {
    if ( $self->wrapped()) {
      $output_file_ref->print($self->format());
    } else {
      $output_file_ref->print($self->string());
    }
    $output_file_ref->print($term);
    $self->clear();
  }
}
