package		OutFile;
require		Exporter;
@ISA =		qw(Exporter);
@EXPORT =	qw(new error name path rename print open close is_open);
# (C) 1999 David C. Fox
# This code may be freely reproduced and modified, provided 
# 1. This copyright notice is retained, 
# and
# 2. substantial modifications are indicated, together with the author's
# name, in any redistributed versions.


use strict;
use English;

use File::Basename;
use FileHandle;

sub new {
  my $invoked = shift;
  my $type = ref( $invoked) || $invoked;
  my $self = {};
  bless $self, $type;
  my $ok;
  if ( $ok = $self->_initialize(@_)) {
    return $self;
  } else {
    return $self->error();
  }
}

sub _initialize {
  my $self = shift;
  my $path = shift;
  $self->{"error"} = "";
  if ( $path) {
    $self->rename($path);
    return $self->open();
  }
  return 1;
}

sub error {
  my $self = shift;
  return $self->{"error"};
}

sub name {
  my $self = shift;
  return $self->{"filename"};
}

sub path {
  my $self = shift;
  return $self->{"path"};
}

sub _output_file_handle {
  my $self = shift;
  return $self->{"file handle"};
}

sub rename {
  my $self = shift;
  my $path = shift;
  if ( $self->is_open()) {
    $self->{"error"} = "file already open";
    return 0;
  }
  $self->{"path"} = $path;
  $self->{"filename"} = basename($path);
  return 1;
}

sub is_open {
  my $self = shift;
  return $self->{"open"};
}

sub open {
  my $self = shift;
  if ( $self->is_open()) {
    $self->{"error"} = "file already open";
    return 0; #  already open
  }

  my $new_name = shift;
  if ( defined($new_name)) {
    $self->rename($new_name);
  }
  my $name = $self->path();

  if (-e  $name) {
    $self->{"error"} = "file $name already exists";
    return 0;
  }
  my $output_handle = new FileHandle $name, "w"; #  open file for writing
  if ( defined $output_handle) {
    $self->{"file handle"} = $output_handle;
    $self->{"open"} = 1;
  } else {
    $self->{"error"} = "opening $name: " . $!;
    return 0;
  }
}

sub close {
  my $self = shift;
  if ( $self->is_open()) {
    my $ok = $self->_output_file_handle()->close();
    if ( $ok) {
      return 1;
    }
    $self->{"error"} = "closing " . $self->path() . $!;
    return 0;
  }
  return 1;
}

sub print {
  my $self = shift;
  my $output_handle = $self->_output_file_handle();
  return $output_handle->print(@_);
}


