package		PhraseTree;
require		Exporter;
require		CopyMe;
@ISA =		qw(CopyMe);
@EXPORT =	qw(new add_phrase complete_match match_phrase);
# (C) 1999 David C. Fox
# This code may be freely reproduced and modified, provided 
# 1. This copyright notice is retained, 
# and
# 2. substantial modifications are indicated, together with the author's
# name, in any redistributed versions.

use strict;
use English;
# tree of translations for phrases.  A phrase is a list of words
# (strings).  A PhraseTree works much like a hash except that it allows
# partial matches on word boundaries and is hierarchical.

# PhraseTree does not perform the task of splitting text string phrases
# into lists of words.

sub new {
  my $invoked = shift;
  my $self;
  my $type = ref( $invoked);
  if ( $type) {
    my %temporary = %$invoked;
    $self = \%temporary;
    bless $self, $type;
  } else {
    $self = {};
    bless $self, $invoked;
    if (!($self->_initialize(@_))) {
      return undef;
    }
  }
  return $self;
}

sub _initialize {
  my $self = shift;
  $self->{"branches"} = {};
  $self->{"value"} = undef;
  return 1;
}

sub add_phrase { #  ($replacement, @phrase)

#  slices off first $word of @phrase
  my ( $self, $replacement, $word, @rest) = @_;

  if ( defined($word)) {
    if ( !(exists($self->{"branches"}{$word}))) {
#       print "no branch $word exists: creating\n";
      $self->{"branches"}{$word} = new PhraseTree();
    }
#     print "adding ", join (" ", @rest), " to branch\n";
    return $self->{"branches"}{$word}->add_phrase($replacement, @rest);
  } else {
    $self->{"value"} = $replacement;
#     print "leaf: value = [$replacement]\n";
    return 1;
  }
}

sub complete_match {#  (@phrase)
#  returns  $replacement, or undef if the phrase does not match
#  completely
  my ( $self, $word, @rest) = @_;

  if ( defined($word)) {
    if ( !(exists ($self->{"branches"}{$word}))) {
      return undef;
    }
    return $self->{"branches"}{$word}->complete_match(@rest);
  } else {
    return $self->{"value"};
  }
}


sub match_phrase {#  (@phrase)
#  returns list of ( $replacement, rest of @phrase unused)
#  $replacement will be undef if there was no match

  my ( $self, $word, @rest) = @_;
  my $replacement;
  my @r;
  if ( defined ($word)) {
    if ( !(exists ($self->{"branches"}{$word}))) {
      return ($self->{"value"}, $word, @rest);
    }
    ($replacement, @r) = $self->{"branches"}{$word}->match_phrase(@rest);
    if ( defined($replacement)) {
      return ($replacement, @r);
    } else {
      return (undef, $word, @rest);
    }
  } else {
    return ($self->{"value"});
  }
}


