package		ScanByLine;
require		Exporter;
@ISA =		qw(Exporter);
@EXPORT =	qw(new eaten error name path rename getline open close is_open
		line_number current_line rest_of_line eof end_of_line
		eat_to_special eat_rest eat_next matches empty_line);
# (C) 1999 David C. Fox
# This code may be freely reproduced and modified, provided 
# 1. This copyright notice is retained, 
# and
# 2. substantial modifications are indicated, together with the author's
# name, in any redistributed versions.

use strict;
use English;

use File::Basename;
use FileHandle;

sub new {
  my $invoked = shift;
  my $type = ref( $invoked) || $invoked;
  my $self = {};
  bless $self, $type;
  my $ok;
  if ( $ok = $self->_initialize(@_)) {
    return $self;
  } else {
    return $self->error();
  }
}

sub _initialize {
#   print "[", @_, "]-";
  my $self = shift;
  my $path = shift;
  my $specials = shift;
  if ( !defined($specials)) {
    $specials = "";
  }
  $self->make_special($specials);
  $self->{"line number"} = 0;
  $self->{"current line"} = "";
  $self->{"rest of line"} = "";
  $self->{"eaten"} = "";
  if ( $path) {
    $self->rename($path);
    return $self->open();
  }
  return 1;
}

sub eaten {
  my $self = shift;
  return $self->{"eaten"};
}

sub daily_specials {
  my $self = shift;
  return $self->{"special characters"};
}

sub make_special {
  my $self = shift;
  my $specials = shift;
  if ( !defined($specials)) {
    $specials = "";
  }
  $self->{"special characters"} = $specials;
}

sub error {
  my $self = shift;
  return $self->{"error"};
}

sub name {
  my $self = shift;
  return $self->{"filename"};
}

sub path {
  my $self = shift;
  return $self->{"path"};
}

sub _input_file_handle {
  my $self = shift;
  return $self->{"file handle"};
}

sub line_number {
  my $self = shift;
  return $self->{"line number"};
}

sub current_line {
  my $self = shift;
  return $self->{"current line"};
}

sub rest_of_line {
  my $self = shift;
  return $self->{"rest of line"};
}

sub rename {
  my $self = shift;
  my $path = shift;
#   print "--$path--\n";
  if ( $self->is_open()) {
    $self->{"error"} = "file already open";
    return 0;
  }
  $self->{"path"} = $path;
  $self->{"filename"} = basename($path);
  return 1;
}

sub is_open {
  my $self = shift;
  return $self->{"open"};
}

sub open {
  my $self = shift;
  if ( $self->is_open()) {
    $self->{"error"} = "file already open";
    return 0; #  already open
  }

  my $new_name = shift;
  if ( defined($new_name)) {
    $self->rename($new_name);
  }
  $self->make_special(shift);

  my $name = $self->path();
  my $input_handle = new FileHandle $name, "r"; #  open file for reading
  if ( defined $input_handle) {
    $self->{"file handle"} = $input_handle;
    $self->{"open"} = 1;
  } else {
    $self->{"error"} = "opening $name: " . $!;
    return 0;
  }
}

sub close {
  my $self = shift;
  if ( $self->is_open()) {
    my $ok = $self->_input_file_handle()->close();
    if ( $ok) {
      return 1;
    }
    $self->{"error"} = "closing " .  $self->path() . $!;
    return 0;
  }
  return 1;
}

sub getline {
  my $self = shift;
  my $input_handle = $self->_input_file_handle();
  my $next_line = $input_handle->getline();
  $self->{"current line"} = $next_line;
  $self->{"rest of line"} = $next_line;
  $self->{"eaten"} = "";
  if ( $next_line) {
    $self->{"line number"} = $input_handle->input_line_number();
  }
  return $next_line;
}

sub eof {
  my $self = shift;
  return eof($self->_input_file_handle());
}

sub end_of_line {
  my $self = shift;
  if ( $self->rest_of_line() eq "") {
    return 1;
  }
  return 0;
}


sub eat_to_special {
  my $self = shift;
  my $specials = shift;
  if ( !defined($specials)) {
    $specials = $self->daily_specials();
     print "default search\n";
  }
  my $search_string = "[$specials\n]";
  if ($self->{"rest of line"} =~ /$search_string/) {
    $self->{"eaten"} .= $` . $&;
#     print "about to eat:\n[$`$&]\n";
    $self->{"rest of line"} = $';
    return ($`, $&);
  }
  my $rest = $self->rest_of_line();
  $self->{"eaten"} .= $rest;
#   print "about to eat rest:\n[$rest]\n";
  $self->{"rest of line"} = "";
  return ($rest, "");
}

sub eat_rest {
  my $self = shift;
  my @value = $self->eat_to_special("");
  return @value;
}

sub matches {
  my $self = shift;
  my $search_string = shift;
  if ( $self->{"rest of line"} =~ /$search_string/) {
    return $&;
  }
  return "";
}

sub eat_next {
  my $self = shift;
  my $what_to_eat = shift;
  my $search_string = "^$what_to_eat";
#   print "-$search_string-\n";
  if ( $self->{"rest of line"} =~ /$search_string/) {
    $self->{"eaten"} .= $&;
    $self->{"rest of line"} = $';
    return $&;
  }
  return "";
}

sub empty_line {
  my $self = shift;
  return ($self->current_line() =~ /^$/);
}


