package		l2tProcess;
require		Exporter;
@ISA =		qw(Exporter);
@EXPORT =	qw(ProcessLatex ProcessDirective LeaveThroughDirective
		ProcessEquation ProcessLatexEquation ProcessDisplayEquation 
		ProcessShortEquation ProcessMacro ProcessEquationMacro 
		);
# (C) 1999 David C. Fox
# This code may be freely reproduced and modified, provided 
# 1. This copyright notice is retained, 
# and
# 2. substantial modifications are indicated, together with the author's
# name, in any redistributed versions.

use strict;
use English;
use ScanLatex ();
use LatexSpecial ();
use EquationSpecial ();
use NatLatexParagraph ();
use PhraseTree ();
use OutFile ();

use ScanErrors;
# use CurrentParagraph;
use newParseLatex;

use WinWait;
use not_dead_yet;


##################################################
# subroutines
##################################################

#############################
# The Parser Subroutines
#############################

#----------------------------
#  the guts of the parser
#----------------------------
sub ProcessLatex {
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
  $directive_marker,  $command_line_options, 
  $macro_tables, $math_macro_tables,
  $group_level) = @_;
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->latex_specials();
  my $extra_term;

  while ( !($input_file_ref->end_of_line()) || !($input_file_ref->eof())) {
    if ($input_file_ref->end_of_line()) {
      #  if end of line, then get a new one
      $input_file_ref->getline();
      #  and check for beginning-of-line specific stuff
      #  (directives)
      if ( $input_file_ref->eat_next($directive_marker) ) {
	ProcessDirective($input_file_ref, $output_file_ref,
	    $paragraph_ref, $directive_marker,  $command_line_options);
      }  elsif ( $input_file_ref->eat_next("\\n") ) {
      #  (or empty lines signifying paragraph breaks)
	if ( $paragraph_ref->empty()) {
	  $paragraph_ref->append("\n");
	  $paragraph_ref->flush_paragraph($output_file_ref, 0);
	} else {
# 	print "empty line";
# 	WarningProcessing($input_file_ref, $paragraph_ref);
	  $paragraph_ref->flush_paragraph($output_file_ref);
	}
      } else {
#         $paragraph_ref->white_space();
        if ( !($paragraph_ref->empty())) {
	  $paragraph_ref->append(" ");
	}
      }
    } else {
    #  if not end of line, then continue processing line
      ($ordinary, $term) = $input_file_ref->eat_to_latex_special();
      $paragraph_ref->append($ordinary);
      if ( $input_file_ref->end_of_line()) {
#         $paragraph_ref->white_space();
      } elsif ( $term eq $special->by_name("tie words")) {
        $paragraph_ref->append($term);
      } elsif ( $term eq $special->by_name("begin comment")) {
        my $comment;
	($comment, $extra_term) = $input_file_ref->eat_rest();
        $paragraph_ref->flush_paragraph($output_file_ref, 1);
        $paragraph_ref->leave($term);
        $paragraph_ref->leave($comment);
	#  add new line and end paragraph to prevent text from next line 
	#  from being wrapped into the comment
        $paragraph_ref->flush_paragraph($output_file_ref, 1);
      } elsif ( $term eq $special->by_name("begin group")) {
        $paragraph_ref->append($term);
        $extra_term = ProcessLatex($input_file_ref, $output_file_ref, 
	    $paragraph_ref, $directive_marker, $command_line_options, 
	    $macro_tables, $math_macro_tables, $group_level + 1);
        $paragraph_ref->append($extra_term);
      } elsif ( $term eq $special->by_name("end group")) {
        if ( $group_level == 0) {
	  print "attempt to end outermost group";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
	return $term;
      } elsif ( $term eq $special->by_name("equation")) {
        if (
	$input_file_ref->eat_next($special->character_named("equation")))
	{
	  #  display equation (TeX style)
	  $paragraph_ref->leave("\\begin{texdisplay}");
	  ProcessDisplayEquation($input_file_ref, $output_file_ref,
	      $paragraph_ref, $directive_marker, $command_line_options, 
	      $macro_tables, $math_macro_tables, 0);
	} else {
	  $paragraph_ref->append('"');
	  ProcessShortEquation($input_file_ref, $output_file_ref,
	      $paragraph_ref, $directive_marker, $command_line_options, 
	      $macro_tables, $math_macro_tables, 0);
	}
      } elsif ( $term eq $special->by_name("macro")) {
        ProcessMacro($input_file_ref, $output_file_ref,
	    $paragraph_ref, $directive_marker, $command_line_options, 
	    $macro_tables, $math_macro_tables, 0);
      } else {
        print "unknown terminator $term";
	ErrorProcessing($input_file_ref, $paragraph_ref);
      }
    }
  }
  #  only one \n
  $paragraph_ref->flush_paragraph($output_file_ref, 1);

  if ( $group_level != 0) {
    print "file ",  $input_file_ref->name();
    print "ended unexpectedly\n";
    print "at group level $group_level != 0";
    ErrorProcessing($input_file_ref, $paragraph_ref);
  }
  if ( $input_file_ref->eof()) {
    return "";
  }

  return $term;
}


#----------------------------
#  directives to the parser to skip certain sections
#----------------------------
sub ProcessDirective{
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
  $directive_marker,  $command_line_options) = @_;
  $paragraph_ref->flush_paragraph($output_file_ref, 1);
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->latex_specials();
  $input_file_ref->eat_next("\\s*");
  if ( $input_file_ref->eat_next("exclude from vocabulary builder") ) {
    $input_file_ref->eat_rest();
    if ( ! ($command_line_options->{"vocabulary builder"})) {
      $output_file_ref->print($input_file_ref->current_line());
    }
    LeaveThroughDirective("end exclude from vocabulary builder", 
	$input_file_ref, $output_file_ref, $paragraph_ref, $directive_marker, 
	$command_line_options);
    return;
  } elsif 
    ( $input_file_ref->eat_next("end exclude from vocabulary builder")) {
    print "directive end exclude from vocabulary builder without\n";
    print "corresponding exclude";
    ErrorProcessing($input_file_ref, $paragraph_ref);
  } else {
    print "unknown directive:", $input_file_ref->rest_of_line();
    ErrorProcessing($input_file_ref, $paragraph_ref);
  }
}

#----------------------------
#  skip sections as instructed 
#----------------------------
sub LeaveThroughDirective{
  my ( $directive, $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker, $command_line_options) = @_;
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->latex_specials();
  while ( !($input_file_ref->end_of_line()) || !($input_file_ref->eof())) {
    if ( !($input_file_ref->end_of_line())) {
      $input_file_ref->eat_rest();
    } else {
      $input_file_ref->getline();
      if ( ! ($command_line_options->{"vocabulary builder"})) {
	$output_file_ref->print($input_file_ref->current_line());
      }
      if ( $input_file_ref->eat_next($directive_marker) ) {
	$input_file_ref->eat_next("\\s*");
	if ( $input_file_ref->eat_next( $directive) ) {
	  $input_file_ref->eat_rest();
	  return;
	} else {
	  print "expecting directive $directive\n";
	  print "not directive:", $input_file_ref->rest_of_line();
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
      } else {
        $input_file_ref->eat_rest();
      }
    }
  }
}




#----------------------------
#   process equation (currently ignored)
#----------------------------
sub ProcessEquation {
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker,  $command_line_options, 
      $macro_tables, $math_macro_tables, $group_level) = @_;
  my $ordinary;
  my $term;
  my $otherterm;
  my $extra_term;
  my $junk;
  my $special = $input_file_ref->equation_specials();
  while ( !($input_file_ref->end_of_line()) || !($input_file_ref->eof())) {
    if ($input_file_ref->end_of_line()) {
      #  if end of line, then get a new one
      $input_file_ref->getline();
      #  and check for beginning-of-line specific stuff
      #  (directives)
      if ( $input_file_ref->eat_next($directive_marker) ) {
        print "no directives allowed within equation mode";
	ErrorProcessing($input_file_ref, $paragraph_ref);
      }  elsif ( $input_file_ref->eat_next("\\n") ) {
	print "no paragraph breaks (blank lines) allowed within ";
        print "equation mode";
	ErrorProcessing($input_file_ref, $paragraph_ref);
      }
    } else {
    #  if not end of line, then continue processing line
      ($ordinary, $term) = $input_file_ref->eat_to_equation_special();
      if ( $ordinary ne "") {
# 	print "[$ordinary]\n";
	$paragraph_ref->math_append($ordinary);
      }
      if ( $input_file_ref->end_of_line()) {
        $paragraph_ref->white_space();
      } elsif ( $term eq $special->by_name("tie words")) {
        $paragraph_ref->append($term);
      } elsif ( $term eq $special->by_name("subscript")) {
        $paragraph_ref->math_append("sub");
      } elsif ( $term eq $special->by_name("super script")) {
	if ( $input_file_ref->eat_next("2")) {
	  $paragraph_ref->math_append("squared");
	} elsif ( $input_file_ref->eat_next("3")) {
	  $paragraph_ref->math_append("cubed");
	} elsif ( $input_file_ref->eat_next("\{-1\}")) {
	  $paragraph_ref->math_append("inverse");
	} elsif ( $input_file_ref->eat_next("\{1/2}")) {
	  $paragraph_ref->math_append("to the one half power");
	} elsif ( $input_file_ref->eat_next("\{-1/2}")) {
	  $paragraph_ref->math_append("to the minus one half power");
	} elsif ( $input_file_ref->eat_next("\{3/2}")) {
	  $paragraph_ref->math_append("to the three halves power");
	} else {
	  $paragraph_ref->math_append("to the power");
	}
      } elsif ( $term eq $special->by_name("plus sign")) {
        $paragraph_ref->math_append("plus");
      } elsif ( $term eq $special->by_name("minus sign")) {
        $paragraph_ref->math_append("minus");
      } elsif ( $term eq $special->by_name("equal sign")) {
        $paragraph_ref->math_append("equals");
      } elsif ( $term eq $special->by_name("greater than")) {
        $paragraph_ref->math_append("greater than");
      } elsif ( $term eq $special->by_name("less than")) {
        $paragraph_ref->math_append("less than");
      } elsif ( $term eq $special->by_name("begin comment")) {
        my $comment;
	($comment, $extra_term) = $input_file_ref->eat_rest();
        $paragraph_ref->leave($term);
        $paragraph_ref->leave("$comment");
	#  add new line and end paragraph to prevent text from next line 
	#  from being wrapped into the comment
	$paragraph_ref->flush_paragraph($output_file_ref, 1);
      } elsif ( $term eq $special->by_name("begin group")) {
	$paragraph_ref->white_space(); 
        $paragraph_ref->append($term);
        ProcessEquation($input_file_ref, $output_file_ref, $paragraph_ref,
	    $directive_marker,  $command_line_options,
	    $macro_tables, $math_macro_tables, $group_level + 1);
      } elsif ( $term eq $special->by_name("end group")) {
        if ( $group_level == 0) {
	  print "attempt to end outermost group in equation mode";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
        $paragraph_ref->math_append($term);
	return;
      } elsif ( $term eq $special->by_name("equation")) {
	return $term;
      } elsif ( $term eq $special->by_name("macro")) {
        $otherterm = ProcessEquationMacro($input_file_ref, $output_file_ref,
	    $paragraph_ref, $directive_marker, $command_line_options,
	    $macro_tables, $math_macro_tables, 0);
	if ( $otherterm ne "") {
	  return $otherterm
	}
      } else {
        print "unknown terminator $term";
	ErrorProcessing($input_file_ref, $paragraph_ref);
      }
    }
  }
  #  only one \n
  $paragraph_ref->flush_paragraph($output_file_ref, 1);

  print "file ended unexpectedly in equation mode\n";
  if ( $group_level != 0) {
    print "at group level $group_level != 0";
  }
  ErrorProcessing($input_file_ref, $paragraph_ref);
}


#----------------------------
#   process display equations (currently ignored)
#----------------------------

sub ProcessLatexEquation {
  my ( $argument, $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker,  $command_line_options, 
      $macro_tables, $math_macro_tables,
      $group_level) = @_;
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->equation_specials();

  $term = ProcessEquation($input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker, $command_line_options,
      $macro_tables, $math_macro_tables,
      $group_level);
  if ( $term eq $special->by_name("equation")) {
    if ( $input_file_ref->eat_next( 
	$special->character_named("equation"))) {
      #  end display equation
#       $paragraph_ref->append('$$');
      print "latex equation \\begin{$argument} ended with \$\$";
      ErrorProcessing($input_file_ref, $paragraph_ref);
    }
    print "latex \\begin{$argument} ended with \$";
    ErrorProcessing($input_file_ref, $paragraph_ref);
  } elsif ( $term eq $argument) {
    $paragraph_ref->flush_paragraph($output_file_ref, 1);
    $paragraph_ref->leave("\\end{$argument}");
    $paragraph_ref->flush_paragraph($output_file_ref, 1);
    return;
  }
  print "unknown terminator $term for latex equation \\begin{$argument}";
  ErrorProcessing($input_file_ref, $paragraph_ref);
}


#----------------------------
#   process display equations (currently ignored)
#----------------------------

sub ProcessDisplayEquation {
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker,  $command_line_options, 
      $macro_tables, $math_macro_tables,
      $group_level) = @_;
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->equation_specials();

  $term = ProcessEquation($input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker, $command_line_options,
      $macro_tables, $math_macro_tables,
      $group_level);
  if ( $term eq $special->by_name("equation")) {
    if ( $input_file_ref->eat_next( 
	$special->character_named("equation"))) {
      #  end display equation
#       $paragraph_ref->append('$$');
      $paragraph_ref->leave("\\end{texdisplay}");
      return;
    }
    print "display equation ended with single \$";
    ErrorProcessing($input_file_ref, $paragraph_ref);
  }
  print "unknown terminator $term for display equation";
  ErrorProcessing($input_file_ref, $paragraph_ref);
}

#----------------------------
#   process inline equations (currently ignored)
#----------------------------

sub ProcessShortEquation {
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker,  $command_line_options, 
      $macro_tables, $math_macro_tables,
      $group_level) = @_;
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->equation_specials();

  $term = ProcessEquation($input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker, $command_line_options,
      $macro_tables, $math_macro_tables,
      $group_level);
  if ( $term eq $special->by_name("equation")) {
    if ( $input_file_ref->eat_next( 
	$special->character_named("equation"))) {
      #  end display equation
      print "in-line equation ended with double \$";
      ErrorProcessing($input_file_ref, $paragraph_ref);
    }
    $paragraph_ref->append('"');
    return;
  }
  print "unknown terminator $term for inline equation";
  ErrorProcessing($input_file_ref, $paragraph_ref);
}

#----------------------------
#   process macros according to instructions
#----------------------------
sub ProcessMacro {
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker,  $command_line_options, 
      $macro_tables, $math_macro_tables,
      $group_level) = @_;
  my $name;
  my $term;
  my $junk;
  my $disposition;
  my $special = $input_file_ref->latex_specials();
  my $argument;
  my $c;
  my $contents;
  my $rest;

  my $extra_term;
  my $replacement;

  my $symbol_ref = $macro_tables->{symbols};
  my $macro_ref = $macro_tables->{macros};
  my $data_ref = $macro_tables->{data};
  my $unknown_macro_file = ${$macro_tables->{unknown}};

  #  next alphabetical string is the name
  $name = DigestControlName($input_file_ref);
  if ( $name) {
    #  macro name
    if ( exists $macro_ref->{$name}) {
      if ( $macro_ref->{$name} eq "begin") {
	$argument = DigestMacroArgument($input_file_ref, $paragraph_ref);
	if ( exists $data_ref->{$name}{$argument}) {
	  $disposition = $data_ref->{$name}{$argument};
	  if ( $disposition eq "leave macro") {
	    $paragraph_ref->leave("\\$name");
	    $paragraph_ref->leave("{$argument}");
	    return;
	  } elsif ( $disposition eq "leave to corresponding end") {
	    $paragraph_ref->leave("\\$name");
	    $paragraph_ref->leave("{$argument}");
	    LeaveThroughMacro("end", $argument, $input_file_ref, 
		$output_file_ref, $paragraph_ref, $directive_marker, 0);
	    return;
	  } elsif ( $disposition eq "equation") {
	    $paragraph_ref->flush_paragraph($output_file_ref, 1);
	    $paragraph_ref->leave("\\$name");
	    $paragraph_ref->leave("{$argument}");
	    #  if $paragraph_ref is set to ignore "leave" arguments, the
	    #  paragraph will still be empty and no extra new line will
	    #  be added by the second flush_paragraph
	    $paragraph_ref->flush_paragraph($output_file_ref, 1);
	    ProcessLatexEquation($argument, $input_file_ref, $output_file_ref,
		$paragraph_ref, $directive_marker, $command_line_options, 
		$macro_tables, $math_macro_tables, 0);
	    return
	  }
	} else {
	  print "\\$name with unknown argument $argument ";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
      } elsif ( $macro_ref->{$name} eq "end") {
	$argument = DigestMacroArgument($input_file_ref, $paragraph_ref);
	if ( exists $data_ref->{$name}{$argument}) {
	  $disposition = $data_ref->{$name}{$argument};
	  if ( $disposition eq "leave macro") {
	    $paragraph_ref->leave("\\$name");
	    $paragraph_ref->leave("{$argument}");
	    return;
	  } elsif ( $disposition eq "leave to corresponding end") {
	    print "leave to corresponding end\{$argument\} without ";
	    print "corresponding begin";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  } elsif ( $disposition eq "equation") {
	    print "equation end\{$argument\} without ";
	    print "corresponding begin";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  } else {
	    print "unknown disposition $disposition ";
	    print "for macro $name\{$argument\}";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} else {
	  print "\\$name with unknown argument $argument ";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
      } else {
	$disposition = $macro_ref->{$name};
	if ( $disposition eq "leave macro") {
	  $paragraph_ref->leave("\\$name");
	} elsif ( $disposition eq "replace") {
 	  $replacement = $data_ref->{$name};
	  if ( ($replacement =~ /^[A-Za-z]/) &&
	       ($paragraph_ref->string() =~ /\a$/)) {
	    print "macro $name with non-delimited replacement\n";
	    print "$replacement must be separated from a preceding\n";
	    print "alphabetical character in text mode by a space.";
	    print "If a space\n";
	    print "is not desired, use a delimited replacement.";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
# 	  $paragraph_ref->white_space_after_most();
 	  $paragraph_ref->append( $data_ref->{$name});
	} elsif ( $disposition eq "leave rest of line") {
	  $paragraph_ref->leave("\\$name");
	  ($junk, $extra_term) = $input_file_ref->eat_rest();
	  $paragraph_ref->leave("$junk");
	  $paragraph_ref->flush_paragraph($output_file_ref, 1);
	} elsif ($disposition eq "parenthetical group") {
	  $junk = $input_file_ref->eat_next("\\s+");
	  $junk = $input_file_ref->eat_next(".");
	  $argument = "";
	  my $auxterm = undef;
	  while ( $junk eq "[") {
	    $argument .= "[";
	    if (!(defined($auxterm))) {
	      ($rest, $auxterm) = $input_file_ref->eat_to_special("\Q{\E");
	    }
#	    print "a--$rest--$auxterm--\n";
	    ($contents, $rest) = DigestBracketed($rest);
#	    print "b--$rest--\n";
	    if ( !(defined($contents))) {
	      print "macro $name with disposition $disposition must\n";
	      print "be followed by an open brace, which could not be found\n";
	      ErrorProcessing($input_file_ref, $paragraph_ref);
	    }
#	    print "$contents--$rest--\n";
	    $argument .= $contents. "]";
	    if ($rest =~ s/^\s*(.)//) {
	      $junk = $&;
	    } else {
	      $junk = $auxterm;
	    }
#	    print "j--$junk--\n";
	  }
	  if ( $junk eq $special->by_name("begin group")) {
	    $paragraph_ref->append("($data_ref->{$name}:");
	    $paragraph_ref->leave($argument);
	    $extra_term = ProcessLatex($input_file_ref, $output_file_ref, 
		$paragraph_ref, $directive_marker, $command_line_options, 
		$macro_tables, $math_macro_tables, 1);
	    $paragraph_ref->append(")");
	  } else {
	    print "macro $name with disposition $disposition must\n";
	    print "be followed by an open brace, not by $junk\n";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} elsif ($disposition eq "make group its own paragraph") {
# 	  print "about to make $name its own paragraph";
# 	  WarningProcessing($input_file_ref, $paragraph_ref);
	  $paragraph_ref->flush_paragraph($output_file_ref, 1);
	  $junk = $input_file_ref->eat_next("\\s+");
	  $junk = $input_file_ref->eat_next(".");
	  if ( $junk eq $special->by_name("begin group")) {
	    $paragraph_ref->leave("\\$name");
	    $paragraph_ref->leave("$junk");
	    $extra_term = ProcessLatex($input_file_ref, $output_file_ref, 
		$paragraph_ref, $directive_marker,  $command_line_options, 
		$macro_tables, $math_macro_tables, 1);
	    $paragraph_ref->leave("$extra_term");
# 	  print "just made $name its own paragraph";
# 	  WarningProcessing($input_file_ref, $paragraph_ref);
	  } else {
	    print "macro $name with disposition $disposition must\n";
	    print "be followed immediately by an open brace, not by $junk\n";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	  $paragraph_ref->flush_paragraph($output_file_ref, 1);
	} elsif ( $disposition eq "leave following group") {
	  $junk = $input_file_ref->eat_next("\\s+");
	  $junk = $input_file_ref->eat_next(".");
	  if ( $junk ne $special->by_name("begin group")) {
	    $paragraph_ref->leave("\\$name");
	    $paragraph_ref->leave("$junk");
	    LeaveGroup($input_file_ref, $output_file_ref, $paragraph_ref,
		$directive_marker, 1);
	    $paragraph_ref->leave("}");
	  } else {
	    print "macro $name with disposition $disposition must\n";
	    print "be followed immediately by an open brace, not by $junk\n";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} elsif ( $disposition eq "leave to macro") {
	  $paragraph_ref->leave("\\$name");
	  LeaveThroughMacro($data_ref->{$name}, "", $input_file_ref, 
	      $output_file_ref, $paragraph_ref, $directive_marker, 0);
	  return;
	} else {
	  print "unknown disposition $disposition ";
	  print "for macro $name";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
      }
    } else {
      if ( $command_line_options->{"ignore"}) {
        $unknown_macro_file->print("\\$name\n");
	$macro_ref->{$name} = "leave macro";
	$paragraph_ref->append("\\$name");
	return;
      }
      print "unknown macro \\$name ";
      ErrorProcessing($input_file_ref, $paragraph_ref);
    }
  } else {
    #  control symbol
    #  next character is the symbol
    $c = $input_file_ref->eat_next(".");
    if ( exists $symbol_ref->{$c}) {
      $paragraph_ref->append($symbol_ref->{$c});
    } else {
      print "unknown control symbol \\$c ";
      ErrorProcessing($input_file_ref, $paragraph_ref);
    }
  }
  return;
}
#----------------------------
#   process macros according to instructions
#----------------------------
#  finish
sub ProcessEquationMacro {
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker,  $command_line_options, 
      $macro_tables, $math_macro_tables, $group_level) = @_;
  my $name;
  my $term;
  my $junk;
  my $disposition;
  my $special = $input_file_ref->latex_specials();
  my $argument;
  my $c;
  my $extra_term;
  my $rest;
  my $contents;


  my $math_symbol_ref = $math_macro_tables->{symbols};
  my $math_macro_ref = $math_macro_tables->{macros};
  my $math_data_ref = $math_macro_tables->{data};
  my $unknown_math_file = ${$math_macro_tables->{unknown}};

  #  next alphabetical string is the name
  $name = DigestControlName($input_file_ref);
  if ( $name) {
    #  macro name
    if ( exists $math_macro_ref->{$name}) {
      if ( $math_macro_ref->{$name} eq "begin") {
	$argument = DigestMacroArgument($input_file_ref, $paragraph_ref);
	if ( exists $math_data_ref->{$name}{$argument}) {
	  $disposition = $math_data_ref->{$name}{$argument};
	  if ( $disposition eq "leave macro") {
	    $paragraph_ref->leave("\\$name");
	    $paragraph_ref->leave("{$argument}");
	    return "";
	  } elsif ( $disposition eq "leave to corresponding end") {
	    $paragraph_ref->leave("\\$name");
	    $paragraph_ref->leave("{$argument}");
	    LeaveThroughMacro("end", $argument, $input_file_ref, 
		$output_file_ref, $paragraph_ref, $directive_marker, 0);
	    return "";
	  } else {
	    print "unknown disposition $disposition for begin";
	    print " $argument in math mode";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} else {
	  print "\\$name with unknown argument $argument ";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
      } elsif ( $math_macro_ref->{$name} eq "end") {
	$argument = DigestMacroArgument($input_file_ref, $paragraph_ref);
	if ( $argument) {
	  if ( exists $math_data_ref->{$name}{$argument}) {
	    $disposition = $math_data_ref->{$name}{$argument};
	    if ( $disposition eq "leave macro") {
	      $paragraph_ref->leave("\\$name");
	      $paragraph_ref->leave("{$argument}");
	      return "";
	    } elsif ( $disposition eq "leave to corresponding end") {
	      print "leave to corresponding end\{$argument\} without ";
	      print "corresponding begin";
	      ErrorProcessing($input_file_ref, $paragraph_ref);
	    } elsif ( $disposition eq "equation") {
	      return $argument;
	    } else {
	      print "unknown disposition $disposition for end";
	      print " $argument in math mode";
	      ErrorProcessing($input_file_ref, $paragraph_ref);
	    }
	  } else {
	    print "\\$name with unknown argument $argument ";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} else {
	  print "missing argument for end notabilities detected";
	  print " by DigestMacroArgument !";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
      } else {
	$disposition = $math_macro_ref->{$name};
	if ( $disposition eq "leave macro") {
	  $paragraph_ref->leave("\\$name");
	} elsif ( $disposition eq "replace") {
	  $paragraph_ref->white_space_after_most(); 
	  #  avoids running words together
	  $paragraph_ref->append( $math_data_ref->{$name});
	} elsif ( $disposition eq "leave rest of line") {
	  $paragraph_ref->leave("\\$name");
	  ($junk, $extra_term) = $input_file_ref->eat_rest();
	  $paragraph_ref->leave($junk);
	  $paragraph_ref->flush_paragraph($output_file_ref, 1);
	} elsif ($disposition eq "parenthetical group") {
	  $junk = $input_file_ref->eat_next("\\s+");
	  $junk = $input_file_ref->eat_next(".");
	  $argument = "";
	  my $auxterm = undef;
	  while ( $junk eq "[") {
	    $argument .= "[";
	    if (!(defined($auxterm))) {
	      ($rest, $auxterm) = $input_file_ref->eat_to_special("\Q{\E");
	    }
#	    print "a--$rest--$auxterm--\n";
	    ($contents, $rest) = DigestBracketed($rest);
#	    print "b--$rest--\n";
	    if ( !(defined($contents))) {
	      print "macro $name with disposition $disposition must\n";
	      print "be followed by an open brace, which could not be found\n";
	      ErrorProcessing($input_file_ref, $paragraph_ref);
	    }
#	    print "$contents--$rest--\n";
	    $argument .= $contents. "]";
	    if ($rest =~ s/^\s*(.)//) {
	      $junk = $&;
	    } else {
	      $junk = $auxterm;
	    }
#	    print "j--$junk--\n";
	  }
	  if ( $junk eq $special->by_name("begin group")) {
	    $paragraph_ref->append("($math_data_ref->{$name}:");
	    $paragraph_ref->leave($argument);
	    $extra_term = ProcessLatex($input_file_ref, $output_file_ref, 
		$paragraph_ref, $directive_marker, $command_line_options, 
		$macro_tables, $math_macro_tables, 1);
	    $paragraph_ref->append(")");
	  } else {
	    print "macro $name with disposition $disposition must\n";
	    print "be followed by an open brace, not by $junk\n";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} elsif ($disposition eq "make group its own paragraph") {
	  print "no paragraph breaks within equation mode ";
	  print "so macro $name with\n";
	  print "disposition $disposition is not allowed";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	} elsif ( $disposition eq "leave following group") {
	  $junk = $input_file_ref->eat_next("\\s+");
	  $junk = $input_file_ref->eat_next(".");
	  if ( $junk ne $special->by_name("begin group")) {
	    $paragraph_ref->leave("\\$name");
	    $paragraph_ref->leave("$junk");
	    LeaveGroup($input_file_ref, $output_file_ref, $paragraph_ref,
		$directive_marker, 1);
	    $paragraph_ref->leave("}");
	  } else {
	    print "macro $name with disposition $disposition must\n";
	    print "be followed immediately by an open brace, not by $junk\n";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} elsif ( $disposition eq "leave to macro") {
	  $paragraph_ref->leave("\\$name");
	  LeaveThroughMacro($math_data_ref->{$name}, "", $input_file_ref, 
	      $output_file_ref, $paragraph_ref, $directive_marker, 0);
	} else {
	  print "unknown disposition $disposition ";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
	return "";
      }
    } else {
      if ( $command_line_options->{"ignore"}) {
        $unknown_math_file->print("\\$name\n");
	$math_macro_ref->{$name} = "leave macro";
	$paragraph_ref->append("\\$name");
	return "";
      }
      print "unknown macro \\$name in math mode";
      ErrorProcessing($input_file_ref, $paragraph_ref);
    }
  } else {
    #  control symbol
    #  next character is the symbol
    $c = $input_file_ref->eat_next(".");
    if ( exists $math_symbol_ref->{$c}) {
      $paragraph_ref->append($math_symbol_ref->{$c});
      return ""
    } else {
      print "unknown control symbol \\$c in math mode";
      ErrorProcessing($input_file_ref, $paragraph_ref);
    }
  }
}

