package		l2tRead;
require		Exporter;
@ISA =		qw(Exporter);
@EXPORT =	qw(ReadMacroDispositions ReadMacroSymbols 
		ReadMacroNames ReadDispositionEatToMacro
		ReadDispositionReplace 
		ValidTranslation ValidParenthetical 
		AddEndDispositions AddEndMathDispositions 
		); 
# (C) 1999 David C. Fox
# This code may be freely reproduced and modified, provided 
# 1. This copyright notice is retained, 
# and
# 2. substantial modifications are indicated, together with the author's
# name, in any redistributed versions.

use strict;
use English;
use ScanLatex ();
use LatexSpecial ();
use NLXSpecial ();
use NatLatexParagraph ();
use PhraseTree ();
use OutFile ();

use ScanErrors;
use CurrentParagraph;
use newParseLatex;

use WinWait;
use not_dead_yet;

#############################
# reading configuration files
#############################

sub ReadMacroDispositions {
# ReadMacroDispositions($filename, $directive_marker, \%macro_symbols,
# \%macro_dispositions, \%macro_data);
  my ( $filename, $directive_marker, $macro_tables) = @_;

  my $file_ref = new ScanLatex($filename);
  ref( $file_ref)  or die_slowly( $file_ref);
  my $special = new LatexSpecial;
  my $directive;
  $file_ref->getline();
  while ( ! ($file_ref->eof()) ) {
    if ( ! ($file_ref->eat_next($directive_marker) )) {
#       print "skipping line:  ",  $file_ref->current_line();
      if ( !($file_ref->eof())) {
	$file_ref->getline();
      }
      next;
    }
#     print "$directive_marker", $file_ref->rest_of_line(),
#     "$directive_marker";
    ($directive) = $file_ref->eat_rest();
    $directive =~ s/^\s*//;
    $directive =~ s/\s*$//;
    if ( $directive  eq "macro symbols") {
      if ( ! (ReadMacroSymbols($special, $directive_marker, $file_ref,
      $macro_tables))) {
        ErrorReading($file_ref);
      }
    } elsif ( $directive eq "macro names") {
      if ( ! (ReadMacroNames($special, $directive_marker, $file_ref,
      $macro_tables))) {
        ErrorReading($file_ref);
      }
    } else {
      print "unrecognized directive ", 
          "$directive_marker $directive $directive_marker\n";
      ErrorReading($file_ref);
    }
  }
}

sub ReadMacroSymbols {
  my ( $special, $directive_marker, $file_ref, $macro_tables) = @_;
  my $c;
  my $translation;
  my $v_translation;
  my $old;
  my $term;
  my $junk;

  my $symbol_ref = $macro_tables->{symbols};

  while ( !( $file_ref->eof())) {
    $file_ref->getline();
    if ( !($file_ref->matches("^$directive_marker"))) {
      #  latex control symbols begin with a backslash.
      #  in the macro disposition file, this can be omitted
      #  (except for \\).  Therefore, we should eat a leading
      #  \  if it is present.
      $file_ref->eat_next($special->character_named("macro"));
      #  next character is the symbol
      $c = $file_ref->eat_next(".");
      if ( exists $symbol_ref->{$c}) {
        print "symbol $c already defined to translate to ";
	print $symbol_ref->{$c}, "\n";
	return 0;
      }
      #  skip following white space
#       print $file_ref->rest_of_line();
      $junk = $file_ref->eat_next("\\s+");
#       print "[$junk]";
#       print $file_ref->rest_of_line();
      if ( $file_ref->eat_next('"')) {
        ($translation, $term) = $file_ref->eat_to_special('"');
	if ( $term ne '"') {
	  print "missing close quotes for symbol \\$c ";
	  return 0;
	}
      } else {
        ($translation) = $file_ref->eat_to_special("\\s");
	if ( $translation eq "") {
	  print "empty translation for symbol \\$c ";
	  return 0;
	}
      }
      $v_translation = ValidTranslation($translation, "\\" . $c, $macro_tables);
      if ( $v_translation) {
	$symbol_ref->{$c} = $v_translation;
      } else {
	print " in replacement string for symbol \\$c ";
        return 0;
      }
#       print "translate $c into [$translation]\n";
    } else {
      return 1;
    }
  }
  return 1;
}


sub ReadMacroNames {
  my ( $special, $directive_marker, $file_ref, $macro_tables) = @_;
  my $name;
  my $disposition;
  my $value;
  my $translation;
  my $argument;

  my $macro_ref = $macro_tables->{macros};
  my $data_ref = $macro_tables->{data};
  my $term;
  my $junk;
  while ( !( $file_ref->eof())) {
    $file_ref->getline();
    if ( !($file_ref->matches("^$directive_marker"))) {
      #  latex control names begin with a backslash.
      #  in the macro disposition file, this can be omitted
      #  (except for \\).  Therefore, we should eat a leading
      #  \  if it is present.
      $file_ref->eat_next($special->character_named("macro"));
      #  next alphabetical string is the name
      $name = DigestControlName($file_ref);
      $junk = $file_ref->eat_next("\\s+");
      if ( $name eq "end") {
        #  special case
        print "disposition for end is determined by that for ";
	print " corresponding begin\n";
	return 0;
      } elsif ( $name eq "begin") {
        #  special case
	if ( !(defined $macro_ref->{$name})) {
	  $macro_ref->{$name} = "begin";
	  $data_ref->{$name} = {};
	}
	$argument = DigestMacroArgument($file_ref);
	$junk = $file_ref->eat_next("\\s+");
	if ( !($file_ref->end_of_line()) ) {
	  if ( $disposition = $file_ref->eat_next("leave macro") ) {
	  } elsif ( $disposition = $file_ref->eat_next("equation") ) {
	  } elsif ($disposition = 
	      $file_ref->eat_next("leave to corresponding end") ) {
	  } else {
	    print "unknown disposition for macro $name\{$argument\} ";
	    return 0;
	  }
	}
# 	print "$name\{$argument\} - [$disposition]\n";
	$data_ref->{$name}{$argument} = $disposition;
      } else {
        
	#  skip following white space
#       print $file_ref->rest_of_line();
	$junk = $file_ref->eat_next("\\s+");
#       print "[$junk]";
#       print $file_ref->rest_of_line();
	if ( !($file_ref->end_of_line()) ) {
	  $value = "";
	  if ( $disposition = $file_ref->eat_next("replace") ) {
	    $translation = ReadDispositionReplace($file_ref, $special);
	    $value = ValidTranslation($translation, "\\" . $name, $macro_tables);
	    if ( ! ($value)) {
	      print " in replacement text for macro \\$name\n";
	      return 0;
	    }
	  } elsif ( $disposition = $file_ref->eat_next("leave macro") ) {
	  } elsif ($disposition = 
	      $file_ref->eat_next("make group its own paragraph") ) {
	  } elsif ($disposition = 
	      $file_ref->eat_next("leave following group") ) {
	  } elsif ( $disposition = $file_ref->eat_next("leave rest of line") ) {
	  } elsif ( $disposition = 
	    $file_ref->eat_next("parenthetical group") ) {
	    $translation = ReadDispositionReplace($file_ref, $special);
	    $value = ValidParenthetical($translation, "\\" . $name, $macro_tables);
	    if ( !($value)) {
	      return 0;
	    }
	  } elsif ( $disposition = $file_ref->eat_next("leave to macro") ) {
	    $value =
		ReadDispositionEatToMacro($file_ref, $special);
	  } else {
	    print "unknown disposition for macro $name ";
	    return 0;
	  }
	}
	if ( exists $macro_ref->{$name}) {
	  print "disposition for macro $name already defined as ";
	  print $macro_ref->{$name}, "\n";
	  return 0;
	}
	$macro_ref->{$name} = $disposition;
# 	print "$name - [$disposition|$value]\n";
	if ( $value) {
	  $data_ref->{$name} = $value;
	}
      }
    } else {
      return 1;
    }
  }
  return 1;
}

sub ReadDispositionEatToMacro {
  my ( $file_ref, $special) = @_;
  my $junk;
  $junk = $file_ref->eat_next("\\s+");
  return $file_ref->eat_next("[A-Za-z]");
}

sub ReadDispositionReplace {
  my ( $file_ref, $special) = @_;
  my $junk;
  my $translation;
  my $term;
  $junk = $file_ref->eat_next("\\s+");
  if ( $file_ref->eat_next('"')) {
    ($translation, $term) = $file_ref->eat_to_special('"');
    if ( $term ne '"') {
      print "missing close quotes ";
    }
  } else {
    ($translation) = $file_ref->eat_to_special("\\s");
  }
#   print "trans:$translation:\n";
  return $translation;
}

sub ValidTranslation { #  ($translation, $macro_tables)
  my ( $translation, $name, $macro_tables) = @_;

  my $phrases = $macro_tables->{phrases};
  my $delimited = $macro_tables->{delimited};
  my $parenthetical = $macro_tables->{parenthetical};

  my $start;
  my $term;
  my $existing;
  my @phrase;

  my $special = new NLXSpecial();



  my $valid = $translation;

  #  check for special characters (not allowed)
  my $specials = $special->character_string();

  if ( $valid =~ /[$specials]/) {
    if ( $valid =~ /^\\/) {
    #  with two exceptions: macros replaced with macros don't generate
    #  t2l replacement strings, so we ignore them
      return $valid;
    }
    #  single special characters
    if ( length($valid) == 1) {
      return $valid;
    }

    print "invalid character ", $&, " in [$translation]\n";
    print "ambiguous with ", $special->by_character($&);
    return "";
  }

  if ( $valid =~ /^\s+/) {
    print "replacement string [$translation] \ncannot begin with white space";
    return "";
  }

  #  check for non-word (alphanumeric or _ ) characters 
  # (indicating a delimited symbol)
  if ( $valid =~ s/^\W+//) {
    $start = $&;
    if ( exists ($delimited->{$start})) {
      $term = $delimited->{$start}{"term"};
      if ( ! ($valid =~ s/\Q$term\E$//)) {
        print "closing delimiter for delimited translation $start\n";
	print "has already been chosen to be [$term].";
	return "";
      }
    } else {
      if ( $valid =~ s/\W+$//) {
        $term = $&;
	$delimited->{$start}{"term"} = $term;
      } else {
        print "delimited translation $start$valid must have a\n";
	print "non-alphanumeric terminator";
	return "";
      }
    }
    if ( $valid =~ /\Q$term\E$/) {
      print "closing delimiter [$term] cannot appear within\n";
      print "replacement string $valid";
      return "";
    }
    if ( ($start eq "(") && ($term eq ":") &&
        exists ($parenthetical->{$valid})) {
      print "delimited translation $start$valid$term is ambiguous\n";
      print "with parenthetical group ";
      print $parenthetical->{$valid};
      return "";
    }
    if ( exists ($delimited->{$start}{"value"}{$valid})) {
      print "delimited translation $start$valid$term is ambiguous\n";
      print "with translation for ";
      print $delimited->{$start}{"value"}{$valid};
      return "";
    }
    $delimited->{$start}{"value"}{$valid} = $name;
    return "$start$valid$term";
  } else {
    if ( $valid =~ /^[A-Za-z\s]+$/) {
      @phrase = split (/\s+/, $valid);
      $valid = join " ", @phrase;
      $existing = $phrases->complete_match(@phrase);
      if ( defined($existing)) {
        print "replacement $valid is ambiguous with replacement for ";
	print $existing;
	return "";
      }
      if ( !($phrases->add_phrase($name, @phrase))) {
        print "failed to add phrase :$valid:\n";
      } else {
	my ( $t, @r) = $phrases->match_phrase(@phrase);
	if ( $name ne $t) {
	  print "failed to recover translation of phrase :$valid:\n";
	}
      }

      return $valid;
    } else {
      print "replacement phrase $valid (without delimiters) cannot\n";
      print "contain non-alphabetical characters except spaces";
      return "";
    }
  }
}

sub ValidParenthetical { #  ($translation, $macro_tables)
  my ( $translation, $name, $macro_tables) = @_;

  my $phrases = $macro_tables->{phrases};
  my $delimited = $macro_tables->{delimited};
  my $parenthetical = $macro_tables->{parenthetical};

  my $start;
  my $term;
  my $existing;
  my @phrase;

  my $special = new NLXSpecial();



  my $valid = $translation;

  #  check for special characters (not allowed)
  my $specials = $special->character_string();

  $start = "(";
  $term = ":";


  if ( $valid =~ /[$specials]/) {
    print "invalid character ", $&, " in [$translation]\n";
    print "ambiguous with ", $special->by_character($&);
    return "";
  }

  if ( $valid =~ /^\s+/) {
    print "replacement string [$translation] \ncannot begin with white space";
    return "";
  }

  if ( $valid =~ /$term/) {
    print "parenthetical group string $valid cannot contain a ':'\n";
    return "";
  }

  if ( exists ($parenthetical->{$valid})) {
    print "parenthetical group for $name is ambiguous with that for\n";
    print $parenthetical->{$valid};
    return "";
  }
  if ( exists ($delimited->{$term})) {
    if ( $delimited->{$start}{"term"} eq $term) {
      if ( exists ($delimited->{$start}{"value"}{$valid})) {
	print "parenthetical group for $name is ambiguous\n";
	print "with translation for ";
	print $delimited->{$start}{"value"}{$valid};
	return "";
      }
    }
  }
  $parenthetical->{$valid} = $name;
#   print "valid parenthetical [$valid]\n";
  return $valid;
}
      
sub AddEndDispositions{
  my ( $macro_tables, $math_macro_tables) = @_;
  my $symbol_ref = $macro_tables->{symbols};
  my $macro_ref = $macro_tables->{macros};
  my $data_ref = $macro_tables->{data};
  my $math_symbol_ref = $math_macro_tables->{symbols};
  my $math_macro_ref = $math_macro_tables->{macros};
  my $math_data_ref = $math_macro_tables->{data};

#   print "does begin exist?\n";
  if ( exists ($macro_ref->{"begin"})) {
    my $argument;
#     print "begin exists\n";
    foreach $argument (keys(%{$data_ref->{"begin"}})) {
      my $disposition = $data_ref->{"begin"}{$argument};
#       print "begin $argument [$disposition]\n";
      if ( $disposition eq "leave to corresponding end") {
	if ( !(exists ($macro_ref->{"end"}))) {
	  $macro_ref->{"end"} = "end";
	  $data_ref->{"end"} = {};
	}
	$data_ref->{"end"}{$argument} = $disposition;
      } elsif ( $disposition eq "equation") {
	if ( !(exists ($math_macro_ref->{"end"}))) {
	  $math_macro_ref->{"end"} = "end";
	  $math_data_ref->{"end"} = {};
	}
	$math_data_ref->{"end"}{$argument} = $disposition;
      } elsif ( $disposition eq "leave macro") {
	if ( !(exists ($macro_ref->{"end"}))) {
	  $macro_ref->{"end"} = "end";
	  $data_ref->{"end"} = {};
	}
	$data_ref->{"end"}{$argument} = $disposition;
      } else {
	die_slowly( "unknown begin/end disposition\n");
      }
    }
  }

#   print "does end exist?\n";
  if ( exists ($macro_ref->{"end"})) {
    my $argument;
#     print "end exists\n";
    foreach $argument (keys(%{$data_ref->{"end"}})) {
      my $disposition = $data_ref->{"end"}{$argument};
#       print "end $argument [$disposition]\n";
    }
  }
}
      
sub AddEndMathDispositions{ 
  my ( $macro_tables, $math_macro_tables) = @_;
  my $symbol_ref = $macro_tables->{symbols};
  my $macro_ref = $macro_tables->{macros};
  my $data_ref = $macro_tables->{data};
  my $math_symbol_ref = $math_macro_tables->{symbols};
  my $math_macro_ref = $math_macro_tables->{macros};
  my $math_data_ref = $math_macro_tables->{data};
#   print "does math begin exist?\n";
  if ( exists ($math_macro_ref->{"begin"})) {
    my $argument;
#     print "math begin exists\n";
    foreach $argument (keys(%{$math_data_ref->{"begin"}})) {
      my $disposition = $math_data_ref->{"begin"}{$argument};
#       print "begin $argument [$disposition]";
      if ( $disposition eq "leave to corresponding end") {
	if ( !(exists ($math_macro_ref->{"end"}))) {
	  $math_macro_ref->{"end"} = "end";
	  $math_data_ref->{"end"} = {};
	}
	$math_data_ref->{"end"}{$argument} = $disposition;
      } elsif ( $disposition eq "leave macro") {
	if ( !(exists ($math_macro_ref->{"end"}))) {
	  $math_macro_ref->{"end"} = "end";
	  $math_data_ref->{"end"} = {};
	}
	$math_data_ref->{"end"}{$argument} = $disposition;
      } else {
	die_slowly( "unknown math mode begin/end disposition\n");
      }
    }
  }


#   print "does math end exist?\n";
  if ( exists ($math_macro_ref->{"end"})) {
    my $argument;
#     print "end exists\n";
    foreach $argument (keys(%{$math_data_ref->{"end"}})) {
      my $disposition = $math_data_ref->{"end"}{$argument};
#       print "end $argument [$disposition]\n";
    }
  }
}
