package		newParseLatex;
require		Exporter;
@ISA =		qw(Exporter);
@EXPORT =	qw(DigestControlName DigestMacroName DigestMacroArgument
		DigestParenthetical
		DigestBracketed
		LeaveBracketed
		FindTrueCharacters FindTrueCharactersAnywhere LeaveMacro
		LeaveThroughMacro LeaveGroup);

use strict;
use English;
# (C) 1999 David C. Fox
# This code may be freely reproduced and modified, provided 
# 1. This copyright notice is retained, 
# and
# 2. substantial modifications are indicated, together with the author's
# name, in any redistributed versions.

use ScanLatex ();
use LatexSpecial ();
use NatLatexParagraph ();
use OutFile ();

use ScanErrors;
use CurrentParagraph;

use not_dead_yet;

#----------------------------
# control names (alphabetical only, + following white space)
#----------------------------
sub DigestControlName {
  my $input_file_ref = shift;
  my $name;
  my $term;
  my $junk;
  #  next alphabetical string is the name
  $name = $input_file_ref->eat_next("[A-Za-z]*");
  if ( $name) {
    #  macro name
#     $junk = $input_file_ref->eat_next("\\s+");
  }
  return $name;
}

#----------------------------
# macro names (alphabetical or single non-alphabetical characters)
#----------------------------
sub DigestMacroName {
  my $input_file_ref = shift;
  my $name;
  my $term;
  my $junk;
  my $c;
  #  next alphabetical string is the name
  $name = $input_file_ref->eat_next("[A-Za-z]*");
  if ( $name) {
    #  macro name
    $junk = $input_file_ref->eat_next("\\s+");
    return $name;
  } else {
    #  control symbol
    #  next character is the symbol
    $c = $input_file_ref->eat_next(".");
    return $c;
  }
}

#----------------------------
#  simple macro argument (in braces)
#----------------------------
sub DigestMacroArgument {
  my ($input_file_ref, $paragraph_ref) = @_;
  my $name;
  my $term;
  my $junk;
  my $argument;
  my $special = $input_file_ref->specials();
  ($junk, $term) = $input_file_ref->eat_to_named( "begin group");
  if ( $term ne $special->by_name("begin group")) {
    print "beginning delimiter is $term\n";
    print "macro $name must have a specified argument ";
    print "enclosed in braces\n";
    ErrorProcessing($input_file_ref, $paragraph_ref);
  }
  ($argument, $term) = $input_file_ref->eat_to_named("end group");
  if ( $term ne $special->by_name("end group")) {
    print "ending delimiter is $term\n";
    print "macro $name must have a specified argument ";
    print "enclosed in braces\n";
    ErrorProcessing($input_file_ref, $paragraph_ref);
  }
  #  this neatens up paragraphs after \begin{abstract} and should be
  #  otherwise harmless
#   $junk = $input_file_ref->eat_next("\\s+");
  return $argument;
}

#----------------------------
#----------------------------
sub DigestParenthetical {
  my ($string) = @_;
  my $name;
  my $term;
  my $junk;
  my $argument = "";
  my $contents;
  my $rest;
  while ( length($string)) {
    if ( $string =~ s/[\(\)]//) {
      $term = $&;
      $argument .= $`; #  text before match
      if ( $term eq ")") {
        return ($argument, $');
      } else {
	$argument .= "(";
        ($contents, $rest) = DigestParenthetical( $string);
	if ( !(defined ($contents))) {
	  $argument .= $contents;
	  $string = $rest;
	}
      }
    } else {
      return (undef, $argument . $string);
    }
  }
}

#----------------------------
#----------------------------
sub DigestBracketed {
  my ($string) = @_;
  my $name;
  my $term;
  my $junk;
  my $argument = "";
  my $contents;
  my $rest;
  while ( length($string)) {
#    print "-$string-\n";
    if ( $string =~ s/[\[\]]//) {
      $term = $&;
      $argument .= $`; #  text before match
#      print "-$term-$argument-$string-\n";
      if ( $term eq "]") {
	# caller doesn't expect final ] to be included
        return ($argument, $');
      } else {
	$argument .= "[";
        ($contents, $rest) = DigestBracketed( $string);
	if ( !(defined ($contents))) {
	  # no matching ] for nested [, so just return 
	  # undef for contents, whole string for rest
	  return (undef, $argument . $string)
	}
	# caller doesn't know about nested [ so include nested ]
	$argument .= $contents . "]";
	$string = $rest;
	# continue looking for outermost ]
      }
    } else {
      # no matching ], so just return undef for contents, whole string
      # for rest
      return (undef, $argument . $string);
    }
  }
  return (undef, $string);
}

#----------------------------
#----------------------------
sub LeaveBracketed {
  my ($input_file_ref, $paragraph_ref) = @_;

  my $name;
  my $term;
  my $extra_term;
  my $junk;
  my @brackets = ("[", "]");

  do {
    $term = FindTrueCharactersAnywhere($input_file_ref, $paragraph_ref,
	@brackets);
    if ( $term eq "]") {
    } elsif ( $term eq "[") {
      $paragraph_ref->leave($term);
      $extra_term = LeaveBracketed($input_file_ref, $paragraph_ref);
      $paragraph_ref->leave($extra_term);
    } else {
      print "file ended unexpectedly while processing LeaveBracketed";
      ErrorProcessing($input_file_ref, $paragraph_ref);
    }
  } while ($term ne "]");

  return $term;
}

#----------------------------
# search within line for specified special character(s), but ignore escaped
# versions (e.g. \$ when searching for $) and contents of comments
#----------------------------
#  returns terminator (specified character if found, otherwise "\n" or
#  "" for end of line/file
sub FindTrueCharacters {
  my ( $input_file_ref, $paragraph_ref, @names) = @_;
  my $target;
  my $term = "";
  my $junk;
  my $special = $input_file_ref->specials();
  my @possible_names = $special->by_character();
  my $name;
  my %possible;
  foreach $name (@possible_names) {
    $possible{$name} = 1;
  }
  my %special_characters;
  foreach $target (@names) {
    if ( exists($possible{$target})) {
      $special_characters{$target} = $special->by_name($target);
    } elsif ( length($target) eq 1) {
      $special_characters{$target} = $target;
    } else {
      print "error: FindTrueCharacters unknown name " . $target;
      ErrorProcessing($input_file_ref, $paragraph_ref);
    }
  }
  while ( !($input_file_ref->end_of_line())) {
    ($junk, $term) = $input_file_ref->eat_to_named(("macro",
	"begin comment", @names));
    $paragraph_ref->leave($junk);
    foreach $target ( @names) {
      if ( $term eq $special_characters{$target}) {
	return $term;
      }
    }
    if ( $term eq $special->by_name("begin comment")) {
      $paragraph_ref->leave($term);
      $junk = $input_file_ref->eat_rest();
      $paragraph_ref->leave($junk);
      return "\n";
    } elsif ( $term eq $special->by_name("macro")) {
      $paragraph_ref->leave($term);
      LeaveMacro($input_file_ref, $paragraph_ref);
    }
  }
  return $term;
}

#----------------------------
# search across lines for specified special character(s), but ignore escaped
# versions (e.g. \$ when searching for $) and contents of comments
#----------------------------
#  returns terminator (specified character if found, otherwise "" 
#  for end of file
sub FindTrueCharactersAnywhere {
  my ( $input_file_ref, $paragraph_ref, @names) = @_;
  my $term;
  do {
    $term = FindTrueCharacters($input_file_ref, $paragraph_ref, @names);
    if ( $term eq "\n") {
      $paragraph_ref->leave($term);
    }
    if ($input_file_ref->end_of_line() ) {
      $input_file_ref->getline();
    } else {
      return $term;
    }
  } while ( !($input_file_ref->eof()) );
  return $term;
}

#----------------------------
#  DigestMacroName and ignore following white space
#----------------------------

sub LeaveMacro {
  my $input_file_ref = shift;
  my $paragraph_ref = shift;
  my $name = DigestMacroName($input_file_ref);
  $paragraph_ref->leave($name);
  my $junk = $input_file_ref->eat_next("\\s+");
  $paragraph_ref->leave($junk);
}

#----------------------------
# ignore everything up to specified macro
#----------------------------
sub LeaveThroughMacro {
  my ( $target_name, $target_argument, $input_file_ref, $output_file_ref, $paragraph_ref,
  $directive_marker, 
       $group_level) = @_;
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->specials();
  my $name;
  my $argument;

  do {
#     print "[Anywhere:target=$target_name, $target_argument]";
#    WarningProcessing( $input_file_ref, $paragraph_ref);
    $term = FindTrueCharactersAnywhere($input_file_ref, 
	$paragraph_ref, "macro");
    if ( $term eq $special->by_name("macro")) {
      $paragraph_ref->leave($term);
      #  next alphabetical string is the name
      $name = DigestMacroName($input_file_ref, $paragraph_ref);
      if ( $name ) {
	$paragraph_ref->leave($name);
	if ( $name eq $target_name) {
	  if ( $target_argument) {
	    $argument = DigestMacroArgument($input_file_ref, $paragraph_ref);
	    $paragraph_ref->leave("{");
	    $paragraph_ref->leave($argument);
	    $paragraph_ref->leave("}");
	    if ( $argument eq $target_argument) {
	      $junk = $input_file_ref->eat_next("\\s+");
	      $paragraph_ref->leave($junk);
	      return;
	    }
	  } else {
	    return;
	  }
	}
      }
    }
  } while ( !($input_file_ref->eof()));
  print "unexpected end of file while eating through macro $target_name,$target_argument";
  ErrorProcessing($input_file_ref, $paragraph_ref);
}

#----------------------------
# ignore contents of rest of current group
#----------------------------
sub LeaveGroup {
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
  $directive_marker, $group_level) = @_;
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->specials();

  while ( !($input_file_ref->end_of_line()) || !($input_file_ref->eof())) {
    $term = FindTrueCharacters($input_file_ref, $paragraph_ref,
	"begin group", "end group");
    if ( $term eq $special->by_name("end group")) {
#       $paragraph_ref->leave("}");
      #  in the case of parenthetical groups, we want to absorb the
      #  closing brace, so we return it so the caller can do what he or
      #  she likes with it.
      return $term;
    } elsif ( $term eq $special->by_name("begin group")) {
      $paragraph_ref->leave("{");
      $paragraph_ref->leave(LeaveGroup($input_file_ref, $output_file_ref,
	  $paragraph_ref, $directive_marker, 0));
    }
    if ( !($input_file_ref->eof())) {
      $input_file_ref->getline();
    }
  }
  print "unexpected end of file while eating group";
  ErrorProcessing($input_file_ref, $paragraph_ref);
}

