#!/opt/bin/perl -w
# (C) 1999 David C. Fox
# This code may be freely reproduced and modified, provided 
# 1. This copyright notice is retained, 
# and
# 2. substantial modifications are indicated, together with the author's
# name, in any redistributed versions.

use strict;
use English;

use File::Basename;
use File::Spec::Functions;

use ScanNLX ();
use NLXSpecial ();
use NLXParagraph ();
use OutFile ();

use ScanErrors;
use CurrentParagraph;
use newParseLatex;
use l2tRead;
use t2lRead;
use t2lProcess;


use WinWait;
use not_dead_yet;



##################################################
# main routine
##################################################

my $directive_marker = "\Q\%NLX\%\E";


my $symbol_ref = {};
my $macro_ref = {};
my $data_ref = {};
my $phrases = new PhraseTree();
my $delimited = {};
my $parenthetical = {};
my $macro_tables = { 
  symbols => $symbol_ref, 
  macros => $macro_ref, 
  data => $data_ref,
  phrases => $phrases,
  delimited => $delimited,
  parenthetical => $parenthetical
};

my $math_symbol_ref = {};
my $math_macro_ref = {};
my $math_data_ref = {};
my $math_phrases = new PhraseTree();
my $math_delimited = {};
my $math_parenthetical = {};
my $math_macro_tables = { 
  symbols => $math_symbol_ref, 
  macros => $math_macro_ref, 
  data => $math_data_ref,
  phrases => $math_phrases,
  delimited => $math_delimited,
  parenthetical => $math_parenthetical
};

my $input_filename;
my %command_line_options;


$input_filename = ReadCommandLine(\%command_line_options);

my ($without_suffix, $path, $suffix) = fileparse($input_filename, "\.txt");
my ($entirely_without_suffix, $no_path, $l_suffix) = 
    fileparse($without_suffix, "\.latex");
$input_filename = "$path$without_suffix" . '.txt';

# find natural latex home directory

my $NaturalLatexHome = $ENV{"NLXHOME"};
if ( !defined($NaturalLatexHome)) {
  die_slowly("NLXHOME not set to NaturalLatex home directory");
}
my ($NLXVolume, $NLXPath) = File::Spec->splitpath($NaturalLatexHome, 1);

#  load translations/dispositions for macros

my $macro_list = File::Spec->catpath($NLXVolume, $NLXPath, "macro.nlx");

ReadMacroDispositions($macro_list, $directive_marker, $macro_tables);

my $specific_macro_list = $path . $entirely_without_suffix . '.macro.nlx';
if (-e $specific_macro_list) {
  ReadMacroDispositions($specific_macro_list, $directive_marker, 
      $macro_tables);
}

AddEndDispositions($macro_tables, $math_macro_tables);

$macro_list = File::Spec->catpath($NLXVolume, $NLXPath, "t2l.nlx");

#  additional translations
ReadTranslations($macro_list, $directive_marker, $macro_tables);

my $math_list = File::Spec->catpath($NLXVolume, $NLXPath, "math.nlx");

ReadMacroDispositions($math_list, $directive_marker, $math_macro_tables);

my $specific_math_list = $path . $entirely_without_suffix . '.math.nlx';
if (-e $specific_math_list) {
  ReadMacroDispositions($specific_math_list, $directive_marker, 
      $math_macro_tables);
}

AddEndMathDispositions($macro_tables, $math_macro_tables);

$math_list = File::Spec->catpath($NLXVolume, $NLXPath, "t2lmath.nlx");

#  additional translations
ReadTranslations($math_list, $directive_marker, $math_macro_tables);

# print "$input_filename]\n";

my $input_file = new ScanNLX($input_filename);
if ( !(ref( $input_file))) {
  die_slowly($input_file);
}

my $output_filename;
my $temporary_filename;
$output_filename = $entirely_without_suffix . '.tex';
$temporary_filename = $entirely_without_suffix . '.working.tex';

if (-e $temporary_filename) {
  unlink ($temporary_filename);
}
my $output_file = new OutFile( $temporary_filename);
if ( !(ref( $output_file))) {
  die_slowly($output_file);
}

my $current_paragraph = new NLXParagraph();

ProcessNLX($input_file, $output_file, $current_paragraph,
    $directive_marker, \%command_line_options, 
    $macro_tables, $math_macro_tables, 0);
$output_file->close();

if (-e $output_filename) {
  my $old_filename = $output_filename . '.old';
  rename $output_filename, $old_filename;
}
rename $temporary_filename, $output_filename;

WinWait();

