package		t2lProcess;
require		Exporter;
@ISA =		qw(Exporter);
@EXPORT =	qw(ProcessNLX ProcessDirective LeaveThroughDirective
		ProcessEquation ProcessNLXEquation 
		ProcessShortEquation ProcessMacro ProcessEquationMacro 
		);
# (C) 1999 David C. Fox
# This code may be freely reproduced and modified, provided 
# 1. This copyright notice is retained, 
# and
# 2. substantial modifications are indicated, together with the author's
# name, in any redistributed versions.

use strict;
use English;
use ScanNLX ();
use NLXSpecial ();
use NLXParagraph ();
use OutFile ();

use ScanErrors;
use newParseLatex;

use WinWait;
use not_dead_yet;


##################################################
# subroutines
##################################################

#############################
# The Parser Subroutines
#############################

#----------------------------
#  the guts of the parser
#----------------------------
sub ProcessNLX {
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
  $directive_marker,  $command_line_options, 
  $macro_tables, $math_macro_tables,
  $group_level) = @_;
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->NLX_specials();
  my $extra_term;

  while ( !($input_file_ref->end_of_line()) || !($input_file_ref->eof())) {
    if ($input_file_ref->end_of_line()) {
      #  if end of line, then get a new one
      $input_file_ref->getline();
      #  and check for beginning-of-line specific stuff
      #  (directives)
      if ( $input_file_ref->eat_next($directive_marker) ) {
	ProcessDirective($input_file_ref, $output_file_ref,
	    $paragraph_ref, $directive_marker,  $command_line_options);
      }  elsif ( $input_file_ref->eat_next("\\n") ) {
      #  (or empty lines signifying paragraph breaks)
	$paragraph_ref->append("\n");
        $paragraph_ref->flush_paragraph($output_file_ref, 1);
      }
    } else {
    #  if not end of line, then continue processing line
      ($ordinary, $term) = $input_file_ref->eat_to_NLX_special();
      ProcessOrdinary($ordinary, $input_file_ref, $paragraph_ref,
	  $macro_tables);
      if ( $input_file_ref->end_of_line()) {
#         $paragraph_ref->append("\n");
        $paragraph_ref->flush_paragraph($output_file_ref, 1);
      } elsif ( $term eq $special->by_name("tie words")) {
        $paragraph_ref->append($term);
      } elsif ( $term eq $special->by_name("dollar sign")) {
        $paragraph_ref->append('\\$');
      } elsif ( $term eq $special->by_name("number sign")) {
        $paragraph_ref->append('\\#');
      } elsif ( $term eq $special->by_name("begin comment")) {
        my $comment;
	($comment, $extra_term) = $input_file_ref->eat_rest();
	$paragraph_ref->flush_paragraph($output_file_ref, 1);
	$paragraph_ref->unwrapped();
        $paragraph_ref->leave($term);
        $paragraph_ref->leave($comment);
	#  add new line and end paragraph to prevent text from next line 
	#  from being wrapped into the comment
	$paragraph_ref->flush_paragraph($output_file_ref, 1);
      } elsif ( $term eq $special->by_name("begin group")) {
        $paragraph_ref->leave($term);
        $extra_term = ProcessNLX($input_file_ref, $output_file_ref, 
	    $paragraph_ref, $directive_marker, $command_line_options, 
	    $macro_tables, $math_macro_tables, $group_level + 1);
        $paragraph_ref->leave($extra_term);
      } elsif ( $term eq $special->by_name("end group")) {
        if ( $group_level == 0) {
	  print "attempt to end outermost group";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
	return $term;
      } elsif ( $term eq $special->by_name("equation")) {
	$paragraph_ref->append('$');
	ProcessShortEquation($input_file_ref, $output_file_ref,
	    $paragraph_ref, $directive_marker, $command_line_options, 
	    $macro_tables, $math_macro_tables, 0);
      } elsif ( $term eq $special->by_name("macro")) {
        ProcessMacro($input_file_ref, $output_file_ref,
	    $paragraph_ref, $directive_marker, $command_line_options, 
	    $macro_tables, $math_macro_tables, 0);
      } else {
        print "unknown terminator $term";
	ErrorProcessing($input_file_ref, $paragraph_ref);
      }
    }
  }
  $paragraph_ref->flush_paragraph($output_file_ref, 1); #  only one \n

  if ( $group_level != 0) {
    print "file ",  $input_file_ref->name();
    print "ended unexpectedly\n";
    print "at group level $group_level != 0";
    ErrorProcessing($input_file_ref, $paragraph_ref);
  }
  if ( $input_file_ref->eof()) {
    return "";
  }

  return $term;
}


#----------------------------
#  directives to the parser to skip certain sections
#----------------------------
sub ProcessDirective{
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
  $directive_marker,  $command_line_options) = @_;
  $paragraph_ref->flush_paragraph($output_file_ref, 1);
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->NLX_specials();
  $input_file_ref->eat_next("\\s*");
  if ( $input_file_ref->eat_next("exclude from vocabulary builder") ) {
    $input_file_ref->eat_rest();
    $output_file_ref->print($input_file_ref->current_line());
    LeaveThroughDirective("end exclude from vocabulary builder", 
	$input_file_ref, $output_file_ref, $paragraph_ref, $directive_marker, 
	$command_line_options);
    return;
  } elsif 
    ( $input_file_ref->eat_next("end exclude from vocabulary builder")) {
    print "directive end exclude from vocabulary builder without\n";
    print "corresponding exclude";
    ErrorProcessing($input_file_ref, $paragraph_ref);
  } else {
    print "unknown directive:", $input_file_ref->rest_of_line();
    ErrorProcessing($input_file_ref, $paragraph_ref);
  }
}

#----------------------------
#  skip sections as instructed 
#----------------------------
sub LeaveThroughDirective{
  my ( $directive, $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker, $command_line_options) = @_;
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->NLX_specials();
  while ( !($input_file_ref->end_of_line()) || !($input_file_ref->eof())) {
    if ( !($input_file_ref->end_of_line())) {
      $input_file_ref->eat_rest();
    } else {
      $input_file_ref->getline();
      $output_file_ref->print($input_file_ref->current_line());
      if ( $input_file_ref->eat_next($directive_marker) ) {
	$input_file_ref->eat_next("\\s*");
	if ( $input_file_ref->eat_next( $directive) ) {
	  $input_file_ref->eat_rest();
	  return;
	} else {
	  print "expecting directive $directive\n";
	  print "not directive:", $input_file_ref->rest_of_line();
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
      } else {
        $input_file_ref->eat_rest();
      }
    }
  }
}




#----------------------------
#   process equation (currently ignored)
#----------------------------
sub ProcessEquation {
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker,  $command_line_options, 
      $macro_tables, $math_macro_tables, $group_level) = @_;
  my $ordinary;
  my $term;
  my $otherterm;
  my $extra_term;
  my $junk;
  my $special = $input_file_ref->NLX_specials();

  $paragraph_ref->equation_mode(1); #use equation mode spacing

  while ( !($input_file_ref->end_of_line()) || !($input_file_ref->eof())) {
    if ($input_file_ref->end_of_line()) {
      #  if end of line, then get a new one
      $input_file_ref->getline();
      #  and check for beginning-of-line specific stuff
      #  (directives)
      if ( $input_file_ref->eat_next($directive_marker) ) {
        print "no directives allowed within equation mode";
	ErrorProcessing($input_file_ref, $paragraph_ref);
      }  elsif ( $input_file_ref->eat_next("\\n") ) {
	print "no paragraph breaks (blank lines) allowed within ";
        print "equation mode";
	ErrorProcessing($input_file_ref, $paragraph_ref);
      }
    } else {
    #  if not end of line, then continue processing line
      ($ordinary, $term) = $input_file_ref->eat_to_NLX_special();
      ProcessOrdinary($ordinary, $input_file_ref, $paragraph_ref,
	  $math_macro_tables);
      if ( $input_file_ref->end_of_line()) {
        $paragraph_ref->flush_paragraph($output_file_ref, 1);
      } elsif ( $term eq $special->by_name("tie words")) {
        $paragraph_ref->append($term);
      } elsif ( $term eq $special->by_name("begin comment")) {
        my $comment;
	($comment, $extra_term) = $input_file_ref->eat_rest();
	$paragraph_ref->flush_paragraph($output_file_ref, 1);
	$paragraph_ref->unwrapped();
        $paragraph_ref->leave($term);
        $paragraph_ref->text_append("$comment");
	#  add new line and end paragraph to prevent text from next line 
	#  from being wrapped into the comment
	$paragraph_ref->flush_paragraph($output_file_ref, 1);
      } elsif ( $term eq $special->by_name("begin group")) {
        $paragraph_ref->leave($term);
        ProcessEquation($input_file_ref, $output_file_ref, $paragraph_ref,
	    $directive_marker,  $command_line_options,
	    $macro_tables, $math_macro_tables, $group_level + 1);
      } elsif ( $term eq $special->by_name("end group")) {
        if ( $group_level == 0) {
	  print "attempt to end outermost group in equation mode";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
        $paragraph_ref->leave($term);
	return;
      } elsif ( $term eq $special->by_name("equation")) {
	$paragraph_ref->equation_mode(0);
	return $term;
      } elsif ( $term eq $special->by_name("macro")) {
        $otherterm = ProcessEquationMacro($input_file_ref, $output_file_ref,
	    $paragraph_ref, $directive_marker, $command_line_options,
	    $macro_tables, $math_macro_tables, 0);
	if ( $otherterm ne "") {
	  $paragraph_ref->equation_mode(0);
	  return $otherterm
	}
      } else {
        print "unknown terminator $term";
	ErrorProcessing($input_file_ref, $paragraph_ref);
      }
    }
  }
  #  only one \n
  $paragraph_ref->flush_paragraph($output_file_ref, 1);

  print "file ended unexpectedly in equation mode\n";
  if ( $group_level != 0) {
    print "at group level $group_level != 0";
  }
  ErrorProcessing($input_file_ref, $paragraph_ref);
}


#----------------------------
#   process display equations 
#----------------------------

sub ProcessNLXEquation {
  my ( $argument, $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker,  $command_line_options, 
      $macro_tables, $math_macro_tables, $group_level) = @_;
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->NLX_specials();

  $term = ProcessEquation($input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker, $command_line_options,
      $macro_tables, $math_macro_tables, $group_level);
  if ( $term eq $special->by_name("equation")) {
    print "latex equation \\begin{$argument} ended with end quotes";
    ErrorProcessing($input_file_ref, $paragraph_ref);
  } elsif ( $term eq $argument) {
    $paragraph_ref->flush_paragraph($output_file_ref, 1);
    $paragraph_ref->leave("\\end{" . $argument . "}");
    $paragraph_ref->flush_paragraph($output_file_ref, 1);
    return;
  }
  print "unknown terminator $term for latex equation \\begin{$argument}";
  ErrorProcessing($input_file_ref, $paragraph_ref);
}

#----------------------------
#   process inline equations (currently ignored)
#----------------------------

sub ProcessShortEquation {
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker,  $command_line_options, 
      $macro_tables, $math_macro_tables, $group_level) = @_;
  my $ordinary;
  my $term;
  my $junk;
  my $special = $input_file_ref->NLX_specials();

  $term = ProcessEquation($input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker, $command_line_options,
      $macro_tables, $math_macro_tables, $group_level);
  if ( $term eq $special->by_name("equation")) {
    $paragraph_ref->append('$');
    return;
  }
  print "unknown terminator $term for inline equation";
  ErrorProcessing($input_file_ref, $paragraph_ref);
}

#----------------------------
#   process macros according to instructions
#----------------------------
sub ProcessMacro {
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker,  $command_line_options, 
      $macro_tables, $math_macro_tables,
      $group_level) = @_;
  my $name;
  my $term;
  my $junk;
  my $disposition;
  my $special = $input_file_ref->NLX_specials();
  my $argument;
  my $c;
  my $rest;
  my $contents;


  my $extra_term;

  my $symbol_ref = $macro_tables->{symbols};
  my $macro_ref = $macro_tables->{macros};
  my $data_ref = $macro_tables->{data};


  #  next alphabetical string is the name
  $name = DigestControlName($input_file_ref);
  if ( $name) {
    #  macro name
    if ( exists $macro_ref->{$name}) {
      if ( $macro_ref->{$name} eq "begin") {
	$argument = DigestMacroArgument($input_file_ref, $paragraph_ref);
	if ( exists $data_ref->{$name}{$argument}) {
	  $disposition = $data_ref->{$name}{$argument};
	  if ( $disposition eq "leave macro") {
	    $paragraph_ref->leave("\\". $name);
	    $paragraph_ref->leave("{$argument}");
	    return;
	  } elsif ( $disposition eq "leave to corresponding end") {
	    $paragraph_ref->leave("\\" . $name);
	    $paragraph_ref->leave("{$argument}");
	    LeaveThroughMacro("end", $argument, $input_file_ref, 
		$output_file_ref, $paragraph_ref, $directive_marker, 0);
	    return;
	  } elsif ( $disposition eq "equation") {
	    $paragraph_ref->flush_paragraph($output_file_ref, 1);
	    $paragraph_ref->append("\\" . $name);
	    $paragraph_ref->append("{$argument}");
	    #  if $paragraph_ref is set to ignore "leave" arguments, the
	    #  paragraph will still be empty and no extra new line will
	    #  be added by the second FlushCurrentParagraph
	    $paragraph_ref->flush_paragraph($output_file_ref, 1);
	    ProcessNLXEquation($argument, $input_file_ref, $output_file_ref,
		$paragraph_ref, $directive_marker, $command_line_options, 
		$macro_tables, $math_macro_tables, 0);
	    return;
	  }
	} else {
	  if ( $command_line_options->{"ignore"}) {
	    print "warning: ";
	    print "\\" . $name . " with unknown argument $argument ";
	    $data_ref->{$name}{$argument} = "leave macro";
	    $data_ref->{"begin"}{$argument} = "leave macro";
	    $paragraph_ref->leave("\\".$name);
	    $paragraph_ref->leave("{$argument}");
	    return;
	  } else {
	    print "\\" . $name. " with unknown argument $argument ";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	}
      } elsif ( $macro_ref->{$name} eq "end") {
	$argument = DigestMacroArgument($input_file_ref, $paragraph_ref);
	if ( exists $data_ref->{$name}{$argument}) {
	  $disposition = $data_ref->{$name}{$argument};
	  if ( $disposition eq "leave macro") {
	    $paragraph_ref->leave("\\".$name);
	    $paragraph_ref->leave("{$argument}");
	    return;
	  } elsif ( $disposition eq "leave to corresponding end") {
	    print "leave to corresponding end\{$argument\} without ";
	    print "corresponding begin";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  } elsif ( $disposition eq "equation") {
	    print "equation end\{$argument\} without ";
	    print "corresponding begin";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  } else {
	    print "unknown disposition $disposition ";
	    print "for macro $name\{$argument\}";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} else {
	  if ( $command_line_options->{"ignore"}) {
	    print "warning: ";
	    print "\\".$name." with unknown argument $argument ";
	    $data_ref->{$name}{$argument} = "leave macro";
	    $data_ref->{"begin"}{$argument} = "leave macro";
	    $paragraph_ref->leave("\\".$name);
	    $paragraph_ref->leave("{$argument}");
	    return;
	  } else {
	    print "\\".$name." with unknown argument $argument ";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	}
      } else {
	$disposition = $macro_ref->{$name};
	if ( $disposition eq "leave macro") {
	  $paragraph_ref->leave("\\".$name);
	} elsif ( $disposition eq "replace") {
	  $paragraph_ref->leave("\\".$name);
	} elsif ( $disposition eq "leave rest of line") {
	  $paragraph_ref->leave("\\".$name);
	  ($junk, $extra_term) = $input_file_ref->eat_rest();
	  $paragraph_ref->leave("$junk");
	  $paragraph_ref->flush_paragraph($output_file_ref, 1);
	} elsif ($disposition eq "parenthetical group") {
	  $junk = $input_file_ref->eat_next("\\s+");
	  $junk = $input_file_ref->eat_next(".");
	  $argument = "";
	  while ( $junk eq "[") {
	    $argument .= "[";
	    $rest = $input_file_ref->eat_to_special("\Q{\E");
	    ($contents, $rest) = DigestBracketed($rest);
	    if ( !(defined($contents))) {
	      print "macro $name with disposition $disposition must\n";
	      print "be followed by an open brace, which could not be found\n";
	      ErrorProcessing($input_file_ref, $paragraph_ref);
	    }
	    $argument .= $contents. "]";
	    $rest =~ s/^\\s*(.)//;
	    $junk = $&;
	  }
	  if ( $junk eq $special->by_name("begin group")) {
	    $paragraph_ref->append("\\".$name);
	    $paragraph_ref->append( $argument);
	    $paragraph_ref->append("\{");
	    $extra_term = ProcessNLX($input_file_ref, $output_file_ref, 
		$paragraph_ref, $directive_marker, $command_line_options, 
		$macro_tables, $math_macro_tables, 1);
	    $paragraph_ref->append("\}");
	  } else {
	    print "macro $name with disposition $disposition must\n";
	    print "be followed by an open brace, not by $junk\n";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} elsif ($disposition eq "make group its own paragraph") {
	  $paragraph_ref->flush_paragraph($output_file_ref);
	  $junk = $input_file_ref->eat_next("\\s+");
	  $junk = $input_file_ref->eat_next(".");
	  if ( $junk eq $special->by_name("begin group")) {
	    $paragraph_ref->leave("\\".$name);
	    $paragraph_ref->leave("$junk");
	    $extra_term = ProcessNLX($input_file_ref, $output_file_ref, 
		$paragraph_ref, $directive_marker,  $command_line_options, 
		$macro_tables, $math_macro_tables, 1);
	    $paragraph_ref->leave("$extra_term");
	  } else {
	    print "macro $name with disposition $disposition must\n";
	    print "be followed immediately by an open brace, not by $junk\n";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	  $paragraph_ref->flush_paragraph($output_file_ref, 1);
	} elsif ( $disposition eq "leave following group") {
	  $junk = $input_file_ref->eat_next("\\s+");
	  $junk = $input_file_ref->eat_next(".");
	  if ( $junk eq $special->by_name("begin group")) {
	    $paragraph_ref->leave("\\".$name);
	    $paragraph_ref->leave("$junk");
	    LeaveGroup($input_file_ref, $output_file_ref, $paragraph_ref,
		$directive_marker, 1);
	    $paragraph_ref->leave("}");
	  } else {
	    print "macro $name with disposition $disposition must\n";
	    print "be followed immediately by an open brace, not by $junk\n";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} elsif ( $disposition eq "leave to macro") {
	  $paragraph_ref->leave("\\" . $name);
	  LeaveThroughMacro($data_ref->{$name}, "", $input_file_ref, 
	      $output_file_ref, $paragraph_ref, $directive_marker, 0);
	  return;
	} else {
	  print "unknown disposition $disposition ";
	  print "for macro $name";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
      }
    } else {
      if ( $command_line_options->{"ignore"}) {
	print "warning: ";
	print "unknown macro \\".$name." ";
	$macro_ref->{$name} = "leave macro";
	$paragraph_ref->leave("\\" . $name);
	return;
      } else {
	print "unknown macro \\" . $name. " ";
	ErrorProcessing($input_file_ref, $paragraph_ref);
      }
    }
  } else {
    #  control symbol
    #  next character is the symbol
    $c = $input_file_ref->eat_next(".");
    if ( exists $symbol_ref->{$c}) {
      $paragraph_ref->append($symbol_ref->{$c});
    } else {
      if ( $command_line_options->{"ignore"}) {
	print "warning: ";
	print "unknown control symbol \\" . $c. " ";
	$macro_ref->{$name} = "leave macro";
	$paragraph_ref->leave("\\" . $c);
	return;
      } else {
	print "unknown control symbol \\" . $c . " ";
	ErrorProcessing($input_file_ref, $paragraph_ref);
      }
    }
  }
  return;
}
#----------------------------
#   process macros according to instructions
#----------------------------
#  finish
sub ProcessEquationMacro {
  my ( $input_file_ref, $output_file_ref, $paragraph_ref,
      $directive_marker,  $command_line_options, 
      $macro_tables, $math_macro_tables, $group_level) = @_;
  my $name;
  my $term;
  my $junk;
  my $disposition;
  my $special = $input_file_ref->NLX_specials();
  my $argument;
  my $c;
  my $extra_term;
  my $rest;
  my $contents;


  my $math_symbol_ref = $math_macro_tables->{symbols};
  my $math_macro_ref = $math_macro_tables->{macros};
  my $math_data_ref = $math_macro_tables->{data};

  #  next alphabetical string is the name
  $name = DigestControlName($input_file_ref);
  if ( $name) {
    #  macro name
    if ( exists $math_macro_ref->{$name}) {
      if ( $math_macro_ref->{$name} eq "begin") {
	$argument = DigestMacroArgument($input_file_ref, $paragraph_ref);
	if ( exists $math_data_ref->{$name}{$argument}) {
	  $disposition = $math_data_ref->{$name}{$argument};
	  if ( $disposition eq "leave macro") {
	    $paragraph_ref->leave("\\" . $name);
	    $paragraph_ref->leave("{$argument}");
	    return "";
	  } elsif ( $disposition eq "leave to corresponding end") {
	    $paragraph_ref->leave("\\" . $name);
	    $paragraph_ref->leave("{$argument}");
	    LeaveThroughMacro("end", $argument, $input_file_ref, 
		$output_file_ref, $paragraph_ref, $directive_marker, 0);
	    return "";
	  } else {
	    print "unknown disposition $disposition for begin";
	    print " $argument in math mode";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} else {
	  print "\\" . $name . " with unknown argument $argument ";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
      } elsif ( $math_macro_ref->{$name} eq "end") {
	$argument = DigestMacroArgument($input_file_ref, $paragraph_ref);
	if ( $argument) {
	  if ( exists $math_data_ref->{$name}{$argument}) {
	    $disposition = $math_data_ref->{$name}{$argument};
	    if ( $disposition eq "leave macro") {
	      $paragraph_ref->leave("\\" . $name);
	      $paragraph_ref->leave("{$argument}");
	      return "";
	    } elsif ( $disposition eq "leave to corresponding end") {
	      print "leave to corresponding end\{$argument\} without ";
	      print "corresponding begin";
	      ErrorProcessing($input_file_ref, $paragraph_ref);
	    } elsif ( $disposition eq "equation") {
	      return $argument;
	    } else {
	      print "unknown disposition $disposition for end";
	      print " $argument in math mode";
	      ErrorProcessing($input_file_ref, $paragraph_ref);
	    }
	  } else {
	    print "\\" . $name . " with unknown argument $argument ";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} else {
	  print "missing argument for end detected";
	  print " by DigestMacroArgument !";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
      } else {
	$disposition = $math_macro_ref->{$name};
	if ( $disposition eq "leave macro") {
	  $paragraph_ref->leave("\\". $name);
	} elsif ( $disposition eq "replace") {
	  $paragraph_ref->append( $math_data_ref->{$name});
	} elsif ( $disposition eq "leave rest of line") {
	  $paragraph_ref->leave("\\".$name);
	  ($junk, $extra_term) = $input_file_ref->eat_rest();
	  $paragraph_ref->leave($junk);
	  $paragraph_ref->flush_paragraph($output_file_ref, 1);
	} elsif ($disposition eq "parenthetical group") {
	  $junk = $input_file_ref->eat_next("\\s+");
	  $junk = $input_file_ref->eat_next(".");
	  $argument = "";
	  while ( $junk eq "[") {
	    $argument .= "[";
	    $rest = $input_file_ref->eat_to_special("\Q{\E");
	    ($contents, $rest) = DigestBracketed($rest);
	    if ( !(defined($contents))) {
	      print "macro $name with disposition $disposition must\n";
	      print "be followed by an open brace, which could not be found\n";
	      ErrorProcessing($input_file_ref, $paragraph_ref);
	    }
	    $argument .= $contents. "]";
	    $rest =~ s/^\\s*(.)//;
	    $junk = $&;
	  }
	  if ( $junk eq $special->by_name("begin group")) {
	    $paragraph_ref->leave("\\".$name);
	    $paragraph_ref->leave( $argument);
	    $paragraph_ref->leave("\{");
	    LeaveGroup($input_file_ref, $output_file_ref, $paragraph_ref,
		$directive_marker, 1);
	    $paragraph_ref->leave("}");
	  } else {
	    print "macro $name with disposition $disposition must\n";
	    print "be followed by an open brace, not by $junk\n";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} elsif ($disposition eq "make group its own paragraph") {
	  print "no paragraph breaks within equation mode ";
	  print "so macro $name with\n";
	  print "disposition $disposition is not allowed";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	} elsif ( $disposition eq "leave following group") {
	  $junk = $input_file_ref->eat_next("\\s+");
	  $junk = $input_file_ref->eat_next(".");
	  if ( $junk eq $special->by_name("begin group")) {
	    $paragraph_ref->leave("\\".$name);
	    $paragraph_ref->leave("$junk");
	    LeaveGroup($input_file_ref, $output_file_ref, $paragraph_ref,
		$directive_marker, 1);
	    $paragraph_ref->leave("}");
	  } else {
	    print "macro $name with disposition $disposition must\n";
	    print "be followed immediately by an open brace, not by $junk\n";
	    ErrorProcessing($input_file_ref, $paragraph_ref);
	  }
	} elsif ( $disposition eq "leave to macro") {
	  $paragraph_ref->leave("\\".$name);
	  LeaveThroughMacro($math_data_ref->{$name}, "", $input_file_ref, 
	      $output_file_ref, $paragraph_ref, $directive_marker, 0);
	} else {
	  print "unknown disposition $disposition ";
	  ErrorProcessing($input_file_ref, $paragraph_ref);
	}
	return "";
      }
    } else {
      print "unknown macro \\".$name." in math mode";
      ErrorProcessing($input_file_ref, $paragraph_ref);
    }
  } else {
    #  control symbol
    #  next character is the symbol
    $c = $input_file_ref->eat_next(".");
    if ( exists $math_symbol_ref->{$c}) {
      $paragraph_ref->append($math_symbol_ref->{$c});
      return ""
    } else {
      print "unknown control symbol \\". $c." in math mode";
      ErrorProcessing($input_file_ref, $paragraph_ref);
    }
  }
}

sub ProcessOrdinary { 
# ($ordinary, $input_file_ref, $paragraph_ref, $macro_tables)
  my ( $ordinary, $input_file_ref, $paragraph_ref, $macro_tables) = @_;

  my $phrases = $macro_tables->{"phrases"};
  my $delimited = $macro_tables->{"delimited"};
  my $parenthetical = $macro_tables->{"parenthetical"};

  my $remaining = $ordinary;
  my $possible;
  my $start;
  my $term;
  my $rest;
  my $contents;
  my $translation;
  my $space = "";



  while ( length($remaining)) {
    if ( $remaining =~ s/^\s+//) {
      $paragraph_ref->append($&);
      next;
    }
    if ( $remaining =~ s/^\d+//) {
      $paragraph_ref->append($&);
      next;
    }
    my $first = substr ( $remaining, 0, 1);
    if ( $remaining =~ /^[^A-Za-z]/) {
      if ( $first eq "(") {
        #  check for parenthetical group
	if ( $remaining =~ /^\((.*?):/) {
	  $possible = $1;
# 	  print "possible parenthetical: $possible\n";
# 	  WarningProcessing($input_file_ref, $paragraph_ref);
	  $rest = $';
	  if ( exists ($parenthetical->{$possible})) {
	    $paragraph_ref->append($parenthetical->{$possible});
	    while ( $rest =~ s/^\s*\[// ) {
	      $paragraph_ref->text_append("[");
	      ($contents, $rest) = DigestBracketed($rest);
	      if ( !(defined($contents))) {
	        print "close bracket for argument to parenthetical";
		print "$possible not found ";
		print "(before next special character) in\n";
		print "[$rest]\n";
		ErrorProcessing($input_file_ref, $paragraph_ref);
	      }
	      $paragraph_ref->text_append($contents . "\]");
	      $remaining = $rest;
	      next;
	    }
	    $paragraph_ref->text_append("\{");
	    ($contents, $rest) = DigestParenthetical($rest);
	    if ( !(defined($contents))) {
	      print "close parenthesis for parenthetical $possible not\n";
	      print "found (before next special character) in\n";
	      print "[$rest]\n";
	      ErrorProcessing($input_file_ref, $paragraph_ref);
	    }
	    $contents =~ s/^\s+//;
	    $paragraph_ref->text_append($contents . "\}");
	    $remaining = $rest;
	    next;
	  }
# 	  print "wasn't though\n";
	  #  if it doesn't turn out to be a parenthetical block, we then
	  #  should check for a delimited replacement starting with "("
	}
	#  however, in that case, "(" is not treated any differently
	#  from any other delimiter, so we just drop down into the
	#  general delimiter case.
      }
      $remaining =~ /^\W+/;
      $start = $&;
      if ( exists ($delimited->{$start})) {
	$term = $delimited->{$start}{"term"};
	if ( $remaining =~ /^\Q$start\E(.*?)\Q$term\E/) {
	  $possible = $1;
	  $rest = $';
	  if ( exists ($delimited->{$start}{"value"}{$possible})) {
	    $paragraph_ref->append(
		$delimited->{$start}{"value"}{$possible});
	    $remaining = $rest;
	    next;
	  }
	  #  not a match to known replacements delimited by $start,
	  #  and $term, so fall through
	}
	#  or expected $term not found
	#  so fall through
      }
      #  or if $start is not a recognized delimiter, then
      #  it doesn't turn out to be a delimited replacement,
      #  so we just eat the first character and try again
      $paragraph_ref->append($first);
      $remaining = substr ( $remaining, 1); 
    } elsif ( $remaining =~ /^[A-Za-z\s]+/) {
    #  doesn't start with white space, or digits, or other 
    # non-alphabetical, so
    #  just check for non-delimited phrases
      $possible = $&;
      my @phrase = split (/\s+/, $possible);
      my @remaining_words;
      ($translation, @remaining_words) = $phrases->match_phrase(@phrase);
      my $used = 1;
      if ( defined($translation)) {
	$paragraph_ref->append($translation);
# 	print "[" . $translation ."]";
	$used = @phrase - @remaining_words;
	while ( $used > 0) {
	  $remaining =~ s/^[A-Za-z]+\s*//;
	  $used--;
	}
      } else {
        $remaining =~ s/^[A-Za-z]+\s*//;
	$paragraph_ref->append($&);
# 	print "<" . $& .">";
      }
    } else {
      print "error in t2lProcess::ProcessOrdinary: don't know how\n";
      print "to handle next character [$first] in\n";
      print "$remaining\n";
      ErrorProcessing($input_file_ref, $paragraph_ref);
    }
  }
}





